# sage.doctest: needs sage.libs.singular
"""
Rational points of curves

We can create points on projective curves::

    sage: P.<x,y,z,w> = ProjectiveSpace(QQ, 3)
    sage: C = Curve([x^3 - 2*x*z^2 - y^3, z^3 - w^3 - x*y*z], P)
    sage: Q = C([1,1,0,0])
    sage: Q.parent()
    Set of rational points of Projective Curve over Rational Field
     defined by x^3 - y^3 - 2*x*z^2, -x*y*z + z^3 - w^3

or on affine curves::

    sage: A.<x,y> = AffineSpace(GF(23), 2)
    sage: C = Curve([y - y^4 + 17*x^2 - 2*x + 22], A)
    sage: Q = C([22,21])
    sage: Q.parent()
    Set of rational points of Affine Plane Curve over Finite Field of size 23
     defined by -y^4 - 6*x^2 - 2*x + y - 1

AUTHORS:

- Grayson Jorgenson (2016-6): initial version
"""
# ****************************************************************************
#       Copyright (C) 2005 William Stein <wstein@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************
from sage.schemes.affine.affine_point import (SchemeMorphism_point_affine_field,
                                              SchemeMorphism_point_affine_finite_field)
from sage.schemes.projective.projective_point import (SchemeMorphism_point_projective_field,
                                                      SchemeMorphism_point_projective_finite_field)


class ProjectiveCurvePoint_field(SchemeMorphism_point_projective_field):
    """
    Point of a projective curve over a field.
    """
    def is_singular(self) -> bool:
        r"""
        Return whether this point is a singular point of the projective curve it is on.

        EXAMPLES::

            sage: P.<x,y,z,w> = ProjectiveSpace(QQ, 3)
            sage: C = Curve([x^2 - y^2, z - w], P)
            sage: Q1 = C([0,0,1,1])
            sage: Q1.is_singular()
            True
            sage: Q2 = C([1,1,1,1])
            sage: Q2.is_singular()
            False
        """
        return self.codomain().is_singular(self)


class ProjectivePlaneCurvePoint_field(ProjectiveCurvePoint_field):
    """
    Point of a projective plane curve over a field.
    """
    def multiplicity(self):
        r"""
        Return the multiplicity of this point with respect to the projective
        curve it is on.

        EXAMPLES::

            sage: P.<x,y,z> = ProjectiveSpace(GF(17), 2)
            sage: C = Curve([y^3*z - 16*x^4], P)
            sage: Q = C([0,0,1])
            sage: Q.multiplicity()
            3
        """
        return self.codomain().multiplicity(self)

    def tangents(self):
        r"""
        Return the tangents at this point of the projective plane curve this point is on.

        OUTPUT:

        A list of polynomials in the coordinate ring of the ambient space
        of the curve this point is on.

        EXAMPLES::

            sage: P.<x,y,z> = ProjectiveSpace(QQ, 2)
            sage: C = Curve([y^2*z^3 - x^5 + 18*y*x*z^3])
            sage: Q = C([0,0,1])
            sage: Q.tangents()
            [y, 18*x + y]
        """
        return self.codomain().tangents(self)

    def is_ordinary_singularity(self) -> bool:
        r"""
        Return whether this point is an ordinary singularity of the projective
        plane curve it is on.

        EXAMPLES::

            sage: P.<x,y,z> = ProjectiveSpace(QQ, 2)
            sage: C = Curve([z^6 - x^6 - x^3*z^3 - x^3*y^3])
            sage: Q = C([0,1,0])
            sage: Q.is_ordinary_singularity()
            False

        ::

            sage: # needs sage.rings.number_field
            sage: R.<a> = QQ[]
            sage: K.<b> = NumberField(a^2 - 3)
            sage: P.<x,y,z> = ProjectiveSpace(K, 2)
            sage: C = P.curve([x^2*y^3*z^4 - y^6*z^3 - 4*x^2*y^4*z^3 -
            ....: 4*x^4*y^2*z^3 + 3*y^7*z^2 + 10*x^2*y^5*z^2 + 9*x^4*y^3*z^2 +
            ....: 5*x^6*y*z^2 - 3*y^8*z - 9*x^2*y^6*z - 11*x^4*y^4*z - 7*x^6*y^2*z -
            ....: 2*x^8*z + y^9 + 2*x^2*y^7 + 3*x^4*y^5 + 4*x^6*y^3 + 2*x^8*y])
            sage: Q = C([-1/2, 1/2, 1])
            sage: Q.is_ordinary_singularity()
            True
        """
        return self.codomain().is_ordinary_singularity(self)

    def is_transverse(self, D) -> bool:
        r"""
        Return whether the intersection of the curve ``D`` at this point with
        the curve this point is on is transverse or not.

        INPUT:

        - ``D`` -- a curve in the same ambient space as the curve this point is on

        EXAMPLES::

            sage: P.<x,y,z> = ProjectiveSpace(QQ, 2)
            sage: C = Curve([x^2 - 2*y^2 - 2*z^2], P)
            sage: D = Curve([y - z], P)
            sage: Q = C([2,1,1])
            sage: Q.is_transverse(D)
            True

        ::

            sage: P.<x,y,z> = ProjectiveSpace(GF(17), 2)
            sage: C = Curve([x^4 - 16*y^3*z], P)
            sage: D = Curve([y^2 - z*x], P)
            sage: Q = C([0,0,1])
            sage: Q.is_transverse(D)
            False
        """
        return self.codomain().is_transverse(D, self)


class ProjectivePlaneCurvePoint_finite_field(ProjectivePlaneCurvePoint_field,
                                             SchemeMorphism_point_projective_finite_field):
    """
    Point of a projective plane curve over a finite field.
    """
    pass


class IntegralProjectiveCurvePoint(ProjectiveCurvePoint_field):
    def closed_point(self):
        """
        Return the closed point corresponding to this rational point.

        EXAMPLES::

            sage: P.<x,y,z> = ProjectiveSpace(GF(17), 2)
            sage: C = Curve([x^4 - 16*y^3*z], P)
            sage: C.singular_points()
            [(0 : 0 : 1)]
            sage: p = _[0]
            sage: p.closed_point()
            Point (x, y)
        """
        curve = self.codomain()
        A = curve.ambient_space()
        S = A.coordinate_ring()

        hcoords = self._coords
        for i in range(S.ngens()):
            if hcoords[i]:
                break
        ai = hcoords[i]
        xi = S.gen(i)
        hgens = [ai * S.gen(j) - hcoords[j] * xi
                 for j in range(S.ngens()) if j != i]
        return curve._closed_point(curve, S.ideal(hgens), degree=1)

    def places(self):
        """
        Return all places on this point.

        EXAMPLES::

            sage: P.<x,y,z> = ProjectiveSpace(GF(17), 2)
            sage: C = Curve([x^4 - 16*y^3*z], P)
            sage: C.singular_points()
            [(0 : 0 : 1)]
            sage: p = _[0]
            sage: p.places()
            [Place (y)]
        """
        return self.closed_point().places()

    def place(self):
        """
        Return a place on this point.

        EXAMPLES::

            sage: P.<x,y,z> = ProjectiveSpace(GF(17), 2)
            sage: C = Curve([x^4 - 16*y^3*z], P)
            sage: C.singular_points()
            [(0 : 0 : 1)]
            sage: p = _[0]
            sage: p.place()
            Place (y)
        """
        return self.closed_point().place()


class IntegralProjectiveCurvePoint_finite_field(IntegralProjectiveCurvePoint):
    """
    Point of an integral projective curve over a finite field.
    """
    pass


class IntegralProjectivePlaneCurvePoint(IntegralProjectiveCurvePoint, ProjectivePlaneCurvePoint_field):
    """
    Point of an integral projective plane curve over a field.
    """
    pass


class IntegralProjectivePlaneCurvePoint_finite_field(ProjectivePlaneCurvePoint_finite_field, IntegralProjectiveCurvePoint_finite_field):
    """
    Point of an integral projective plane curve over a finite field.
    """
    pass


class AffineCurvePoint_field(SchemeMorphism_point_affine_field):

    def is_singular(self) -> bool:
        r"""
        Return whether this point is a singular point of the affine curve it is on.

        EXAMPLES::

            sage: # needs sage.rings.number_field
            sage: K = QuadraticField(-1)
            sage: A.<x,y,z> = AffineSpace(K, 3)
            sage: C = Curve([(x^4 + 2*z + 2)*y, z - y + 1])
            sage: Q1 = C([0,0,-1])
            sage: Q1.is_singular()
            True
            sage: Q2 = C([-K.gen(),0,-1])
            sage: Q2.is_singular()
            False
        """
        return self.codomain().is_singular(self)


class AffinePlaneCurvePoint_field(AffineCurvePoint_field):
    """
    Point of an affine plane curve over a field.
    """
    def multiplicity(self):
        r"""
        Return the multiplicity of this point with respect to the affine curve it is on.

        EXAMPLES::

            sage: A.<x,y> = AffineSpace(QQ, 2)
            sage: C = A.curve([2*x^7 - 3*x^6*y + x^5*y^2 + 31*x^6 - 40*x^5*y +
            ....: 13*x^4*y^2 - x^3*y^3 + 207*x^5 - 228*x^4*y + 70*x^3*y^2 - 7*x^2*y^3
            ....: + 775*x^4 - 713*x^3*y + 193*x^2*y^2 - 19*x*y^3 + y^4 + 1764*x^3 -
            ....: 1293*x^2*y + 277*x*y^2 - 22*y^3 + 2451*x^2 - 1297*x*y + 172*y^2 +
            ....: 1935*x - 570*y + 675])
            sage: Q = C([-2,1])
            sage: Q.multiplicity()
            4
        """
        return self.codomain().multiplicity(self)

    def tangents(self):
        r"""
        Return the tangents at this point of the affine plane curve this point is on.

        OUTPUT: list of polynomials in the coordinate ring of the ambient
        space of the curve this point is on

        EXAMPLES::

            sage: A.<x,y> = AffineSpace(QQ, 2)
            sage: C = A.curve([x^5 - x^3*y^2 + 5*x^4 - x^3*y - 3*x^2*y^2 +
            ....: x*y^3 + 10*x^3 - 3*x^2*y - 3*x*y^2 + y^3 + 10*x^2 - 3*x*y - y^2 +
            ....: 5*x - y + 1])
            sage: Q = C([-1,0])
            sage: Q.tangents()
            [y, x - y + 1, x + 1, x + y + 1]
        """
        return self.codomain().tangents(self)

    def is_ordinary_singularity(self) -> bool:
        r"""
        Return whether this point is an ordinary singularity of the affine
        plane curve it is on.

        EXAMPLES::

            sage: A.<x,y> = AffineSpace(QQ, 2)
            sage: C = A.curve([x^5 - x^3*y^2 + 5*x^4 - x^3*y - 3*x^2*y^2 +
            ....: x*y^3 + 10*x^3 - 3*x^2*y - 3*x*y^2 + y^3 + 10*x^2 - 3*x*y - y^2 +
            ....: 5*x - y + 1])
            sage: Q = C([-1,0])
            sage: Q.is_ordinary_singularity()
            True

        ::

            sage: A.<x,y> = AffineSpace(GF(7), 2)
            sage: C = A.curve([y^2 - x^7 - 6*x^3])
            sage: Q = C([0,0])
            sage: Q.is_ordinary_singularity()
            False
        """
        return self.codomain().is_ordinary_singularity(self)

    def is_transverse(self, D) -> bool:
        r"""
        Return whether the intersection of the curve ``D`` at this point with
        the curve this point is on is transverse or not.

        INPUT:

        - ``D`` -- a curve in the same ambient space as the curve this point is on

        EXAMPLES::

            sage: A.<x,y> = AffineSpace(QQ, 2)
            sage: C = Curve([y - x^2], A)
            sage: D = Curve([y], A)
            sage: Q = C([0,0])
            sage: Q.is_transverse(D)
            False

        ::

            sage: # needs sage.rings.number_field
            sage: R.<a> = QQ[]
            sage: K.<b> = NumberField(a^2 - 2)
            sage: A.<x,y> = AffineSpace(K, 2)
            sage: C = Curve([y^2 + x^2 - 1], A)
            sage: D = Curve([y - x], A)
            sage: Q = C([-1/2*b, -1/2*b])
            sage: Q.is_transverse(D)
            True
        """
        return self.codomain().is_transverse(D, self)


class AffinePlaneCurvePoint_finite_field(AffinePlaneCurvePoint_field, SchemeMorphism_point_affine_finite_field):
    """
    Point of an affine plane curve over a finite field.
    """
    pass


class IntegralAffineCurvePoint(AffineCurvePoint_field):
    """
    Point of an integral affine curve.
    """
    def closed_point(self):
        """
        Return the closed point that corresponds to this rational point.

        EXAMPLES::

            sage: # needs sage.rings.finite_rings
            sage: A.<x,y> = AffineSpace(GF(8), 2)
            sage: C = Curve(x^5 + y^5 + x*y + 1)
            sage: p = C([1,1])
            sage: p.closed_point()
            Point (x + 1, y + 1)
        """
        curve = self.codomain()
        A = curve.ambient_space()
        R = A.coordinate_ring()
        coords = self._coords
        gens = [R.gen(i) - coords[i] for i in range(R.ngens())]
        return curve._closed_point(curve, R.ideal(gens), degree=1)

    def places(self):
        """
        Return all places on this point.

        EXAMPLES::

            sage: A.<x,y> = AffineSpace(GF(2), 2)
            sage: C = Curve(x^5 + y^5 + x*y + 1)
            sage: p = C(-1,-1)
            sage: p
            (1, 1)
            sage: p.closed_point()
            Point (x + 1, y + 1)
            sage: _.places()
            [Place (x + 1, (1/(x^5 + 1))*y^4 + ((x^5 + x^4 + 1)/(x^5 + 1))*y^3
                           + ((x^5 + x^3 + 1)/(x^5 + 1))*y^2 + (x^2/(x^5 + 1))*y),
             Place (x + 1, (1/(x^5 + 1))*y^4 + ((x^5 + x^4 + 1)/(x^5 + 1))*y^3
                           + (x^3/(x^5 + 1))*y^2 + (x^2/(x^5 + 1))*y + x + 1)]
        """
        return self.closed_point().places()

    def place(self):
        """
        Return a place on this point.

        EXAMPLES::

            sage: A.<x,y> = AffineSpace(GF(2), 2)
            sage: C = Curve(x^5 + y^5 + x*y + 1)
            sage: p = C(-1,-1)
            sage: p
            (1, 1)
            sage: p.closed_point()
            Point (x + 1, y + 1)
            sage: _.place()
            Place (x + 1, (1/(x^5 + 1))*y^4 + ((x^5 + x^4 + 1)/(x^5 + 1))*y^3
                          + ((x^5 + x^3 + 1)/(x^5 + 1))*y^2 + (x^2/(x^5 + 1))*y)
        """
        return self.closed_point().place()


class IntegralAffineCurvePoint_finite_field(IntegralAffineCurvePoint):
    """
    Point of an integral affine curve over a finite field.
    """
    pass


class IntegralAffinePlaneCurvePoint(IntegralAffineCurvePoint, AffinePlaneCurvePoint_field):
    """
    Point of an integral affine plane curve.
    """
    pass


class IntegralAffinePlaneCurvePoint_finite_field(AffinePlaneCurvePoint_finite_field, IntegralAffineCurvePoint_finite_field):
    """
    Point of an integral affine plane curve over a finite field.
    """
    pass



# sage.doctest: needs sage.libs.gap
r"""
Overview of Hecke triangle groups and modular forms for Hecke triangle groups

AUTHORS:

- Jonas Jermann (2013): initial version


Hecke triangle groups and elements:
-----------------------------------

- **Hecke triangle group:**
  The Von Dyck group corresponding to the triangle group with angles
  ``(pi/2, pi/n, 0)`` for ``n=3, 4, 5, ...``, generated by the conformal
  circle inversion ``S`` and by the translation ``T`` by ``lambda=2*cos(pi/n)``.
  I.e. the subgroup of orientation preserving elements of the triangle
  group generated by reflections along the boundaries of the above hyperbolic
  triangle. The group is arithmetic iff ``n=3, 4, 6, infinity``.

  The group elements correspond to matrices over ZZ[lambda], namely the
  corresponding order in the number field defined by the minimal polynomial
  of lambda (which embeds into ``AlgebraicReal`` accordingly).

  An exact symbolic expression of the corresponding transfinite diameter ``d``
  (which is used as a formal parameter for Fourier expansion of modular forms)
  can be obtained. For arithmetic groups the (correct) rational number is
  returned instead.

  Basic matrices like ``S, T, U, V(j)`` are available.

  EXAMPLES::

      sage: from sage.modular.modform_hecketriangle.hecke_triangle_groups import HeckeTriangleGroup
      sage: G = HeckeTriangleGroup(12)
      sage: G
      Hecke triangle group for n = 12
      sage: G.is_arithmetic()
      False
      sage: G.dvalue()
      e^(2*euler_gamma - 4*pi/(sqrt(6) + sqrt(2)) + psi(19/24) + psi(17/24))
      sage: AA(G.lam())
      1.9318516525781...?

      sage: G = HeckeTriangleGroup(6)
      sage: G
      Hecke triangle group for n = 6
      sage: G.is_arithmetic()
      True
      sage: G.dvalue()
      1/108
      sage: AA(G.lam()) == AA(sqrt(3))
      True
      sage: G.gens()
      (
      [ 0 -1]  [  1 lam]
      [ 1  0], [  0   1]
      )
      sage: G.U()^3
      [ lam   -2]
      [   2 -lam]
      sage: G.U().parent()
      Hecke triangle group for n = 6
      sage: G.U().matrix().parent()
      Full MatrixSpace of 2 by 2 dense matrices over Maximal Order generated by lam in Number Field in lam with defining polynomial x^2 - 3 with lam = 1.732050807568878?


- **Decomposition into product of generators:**
  It is possible to decompose any group element into products
  of generators the ``S`` and ``T``. In particular this allows
  to check whether a given matrix indeed is a group element.

  It also allows one to calculate the automorphy factor
  of a modular form for the Hecke triangle group for
  arbitrary arguments.

  EXAMPLES::

      sage: from sage.modular.modform_hecketriangle.hecke_triangle_groups import HeckeTriangleGroup
      sage: G = HeckeTriangleGroup(6)
      sage: G.element_repr_method("basic")
      sage: A = G.V(2)*G.V(3)^(-2)
      sage: (L, sgn) = A.word_S_T()
      sage: L
      (S, T^(-2), S, T^(-1), S, T^(-1))
      sage: sgn
      -1
      sage: sgn.parent()
      Hecke triangle group for n = 6

      sage: G(matrix([[-1, 1+G.lam()],[0, -1]]))
      Traceback (most recent call last):
      ...
      TypeError: The matrix is not an element of Hecke triangle group for n = 6, up to equivalence it identifies two nonequivalent points.
      sage: G(matrix([[-1, G.lam()],[0, -1]]))
      -T^(-1)

      sage: G.element_repr_method("basic")

      sage: from sage.modular.modform_hecketriangle.space import ModularForms
      sage: MF = ModularForms(G, k=4, ep=1)
      sage: z = AlgebraicField()(1+i/2)
      sage: MF.aut_factor(A, z)
      37.62113890008...? + 12.18405525839...?*I


- **Representation of elements:**
  An element can be represented in several ways:

  - As a matrix over the base ring (default)
  - As a product of the generators ``S`` and ``T``
  - As a product of basic blocks conjugated by some element

  EXAMPLES::

      sage: from sage.modular.modform_hecketriangle.hecke_triangle_groups import HeckeTriangleGroup
      sage: G = HeckeTriangleGroup(n=5)
      sage: el = G.S()*G.T(3)*G.S()*G.T(-2)

      sage: G.element_repr_method("default")
      sage: el
      [        -1      2*lam]
      [     3*lam -6*lam - 7]

      sage: G.element_repr_method("basic")
      sage: el
      S*T^3*S*T^(-2)

      sage: G.element_repr_method("block")
      sage: el
      -(S*T^3) * (V(4)^2*V(1)^3) * (S*T^3)^(-1)

      sage: G.element_repr_method("conj")
      sage: el
      [-V(4)^2*V(1)^3]

      sage: G.element_repr_method("default")


- **Group action on the (extended) upper half plane:**
  The group action of Hecke triangle groups on the (extended)
  upper half plane (by linear fractional transformations)
  is implemented. The implementation is not based on a
  specific upper half plane model but is defined formally
  (for arbitrary arguments) instead.

  It is possible to determine the group translate of an
  element in the classic (strict) fundamental domain for
  the group, together with the corresponding mapping
  group element.

  The corresponding action of the group on itself by
  conjugation is supported as well.

  The usual `slash`-operator for even integer weights is
  also available. It acts on rational functions (resp. polynomials).
  For modular forms an evaluation argument is required.

  EXAMPLES::

      sage: from sage.modular.modform_hecketriangle.hecke_triangle_groups import HeckeTriangleGroup
      sage: G = HeckeTriangleGroup(n=7)
      sage: G.element_repr_method("basic")

      sage: G.S().acton(i + exp(-2))
      -1/(e^(-2) + I)
      sage: A = G.V(2)*G.V(3)^(-2)
      sage: A
      -S*T^(-2)*S*T^(-1)*S*T^(-1)
      sage: A.acton(CC(i + exp(-2)))
      0.344549645079... + 0.0163901095115...*I

      sage: G.S().acton(A)
      -T^(-2)*S*T^(-1)*S*T^(-1)*S

      sage: z = AlgebraicField()(4 + 1/7*i)
      sage: G.in_FD(z)
      False
      sage: A, w = G.get_FD(z)
      sage: A
      T^2*S*T^(-1)*S
      sage: w
      0.516937798396...? + 0.964078044600...?*I

      sage: A.acton(w) == z
      True
      sage: G.in_FD(w)
      True

      sage: z = PolynomialRing(G.base_ring(), 'z').gen()
      sage: rat = z^2 + 1/(z-G.lam())
      sage: G.S().slash(rat)
      (z^6 - lam*z^4 - z^3)/(-lam*z^4 - z^3)

      sage: G.element_repr_method("default")


- **Basic properties of group elements:**
  The trace, sign (based on the trace), discriminant and
  elliptic/parabolic/hyperbolic type are available.

  Group elements can be displayed/represented in several ways:

  - As matrices over the base ring.
  - As a word in (powers of) the generators ``S`` and ``T``.
  - As a word in (powers of) basic block matrices ``V(j)``
    (resp. ``U, S`` in the elliptic case) together
    with the conjugation matrix that maps the element to
    this form (also see below).

  For the case ``n=infinity`` the last method is not
  properly implemented.

  EXAMPLES::

      sage: from sage.modular.modform_hecketriangle.hecke_triangle_groups import HeckeTriangleGroup
      sage: G = HeckeTriangleGroup(n=7)
      sage: A = -G.V(2)*G.V(3)^(-2)

      sage: print(A.string_repr("default"))
      [               lam         -lam^2 + 1]
      [       2*lam^2 - 1 -2*lam^2 - lam + 2]
      sage: print(A.string_repr("basic"))
      S*T^(-2)*S*T^(-1)*S*T^(-1)
      sage: print(A.string_repr("block"))
      -(-S*T^(-1)*S) * (V(3)) * (-S*T^(-1)*S)^(-1)
      sage: print(A.string_repr("conj"))
      [-V(3)]
      sage: A.trace()
      -2*lam^2 + 2
      sage: A.sign()
      [-1  0]
      [ 0 -1]
      sage: A.discriminant()
      4*lam^2 + 4*lam - 4
      sage: A.is_elliptic()
      False
      sage: A.is_hyperbolic()
      True


- **Fixed points:**
  Elliptic, parabolic or hyperbolic fixed points of group
  can be obtained. They are implemented as a (relative)
  quadratic extension (given by the square root of the discriminant)
  of the base ring. It is possible to query the correct embedding
  into a given field.

  Note that for hyperbolic (and parabolic) fixed points there is a
  1-1 correspondence with primitive hyperbolic/parabolic group
  elements (at least if ``n < infinity``). The group action on
  fixed points resp. on matrices is compatible with this correspondence.

  EXAMPLES::

      sage: from sage.modular.modform_hecketriangle.hecke_triangle_groups import HeckeTriangleGroup
      sage: G = HeckeTriangleGroup(n=7)

      sage: A = G.S()
      sage: A.fixed_points()
      (1/2*e, -1/2*e)
      sage: A.fixed_points(embedded=True)
      (I, -I)

      sage: A = G.U()
      sage: A.fixed_points()
      (1/2*e + 1/2*lam, -1/2*e + 1/2*lam)
      sage: A.fixed_points(embedded=True)
      (0.9009688679024...? + 0.4338837391175...?*I, 0.9009688679024...? - 0.4338837391175...?*I)

      sage: A = -G.V(2)*G.V(3)^(-2)
      sage: A.fixed_points()
      ((-3/7*lam^2 + 2/7*lam + 11/14)*e - 1/7*lam^2 + 3/7*lam + 3/7, (3/7*lam^2 - 2/7*lam - 11/14)*e - 1/7*lam^2 + 3/7*lam + 3/7)
      sage: A.fixed_points(embedded=True)
      (0.3707208390178...?, 1.103231619181...?)

      sage: el = A.fixed_points()[0]
      sage: F = A.root_extension_field()
      sage: F == el.parent()
      True
      sage: A.root_extension_embedding(CC)
      Relative number field morphism:
        From: Number Field in e with defining polynomial x^2 - 4*lam^2 - 4*lam + 4 over its base field
        To:   Complex Field with 53 bits of precision
        Defn: e |--> 4.02438434522465
              lam |--> 1.80193773580484

      sage: G.V(2).acton(A).fixed_points()[0] == G.V(2).acton(el)
      True


- **Lambda-continued fractions:**
  For parabolic or hyperbolic elements (resp. their corresponding
  fixed point) the (negative) lambda-continued fraction expansion
  is eventually periodic. The lambda-CF (i.e. the preperiod and period)
  is calculated exactly.

  In particular this allows to determine primitive and reduced
  generators of group elements and the corresponding primitive
  power of the element.

  The case ``n=infinity`` is not properly implemented.

  EXAMPLES::

      sage: from sage.modular.modform_hecketriangle.hecke_triangle_groups import HeckeTriangleGroup
      sage: G = HeckeTriangleGroup(n=7)
      sage: G.element_repr_method("block")

      sage: G.V(6).continued_fraction()
      ((1,), (1, 1, 1, 1, 2))
      sage: (-G.V(2)).continued_fraction()
      ((1,), (2,))

      sage: A = -(G.V(2)*G.V(3)^(-2))^2
      sage: A.is_primitive()
      False
      sage: A.primitive_power()
      2
      sage: A.is_reduced()
      False
      sage: A.continued_fraction()
      ((1, 1, 1, 1), (1, 2))

      sage: B = A.primitive_part()
      sage: B
      (-S*T^(-1)*S) * (V(3)) * (-S*T^(-1)*S)^(-1)
      sage: B.is_primitive()
      True
      sage: B.is_reduced()
      False
      sage: B.continued_fraction()
      ((1, 1, 1, 1), (1, 2))
      sage: A == A.sign() * B^A.primitive_power()
      True

      sage: B = A.reduce()
      sage: B
      (T*S*T) * (V(3)) * (T*S*T)^(-1)
      sage: B.is_primitive()
      True
      sage: B.is_reduced()
      True
      sage: B.continued_fraction()
      ((), (1, 2))

      sage: G.element_repr_method("default")


- **Reduced and simple elements, Hecke-symmetric elements:**
  For primitive conjugacy classes of hyperbolic elements
  the cycle of reduced elements can be obtain as well as
  all simple elements. It is also possible to determine
  whether a class is Hecke-symmetric.

  The case ``n=infinity`` is not properly implemented.

  EXAMPLES::

      sage: from sage.modular.modform_hecketriangle.hecke_triangle_groups import HeckeTriangleGroup
      sage: G = HeckeTriangleGroup(n=5)

      sage: el = G.V(1)^2*G.V(2)*G.V(4)
      sage: R = el.reduced_elements()
      sage: [v.continued_fraction() for v in R]
      [((), (2, 1, 1, 4)), ((), (1, 1, 4, 2)), ((), (1, 4, 2, 1)), ((), (4, 2, 1, 1))]

      sage: el = G.V(1)^2*G.V(2)*G.V(4)
      sage: R = el.simple_elements()
      sage: [v.is_simple() for v in R]
      [True, True, True, True]
      sage: (fp1, fp2) = R[2].fixed_points(embedded=True)
      sage: fp2 < 0 < fp1
      True

      sage: el = G.V(2)
      sage: el.is_hecke_symmetric()
      False
      sage: (el.simple_fixed_point_set(), el.inverse().simple_fixed_point_set())
      ({1/2*e, (-1/2*lam + 1/2)*e}, {-1/2*e, (1/2*lam - 1/2)*e})
      sage: el = G.V(2)*G.V(3)
      sage: el.is_hecke_symmetric()
      True
      sage: el.simple_fixed_point_set() == el.inverse().simple_fixed_point_set()
      True


- **Rational period functions:**
  For each primitive (hyperbolic) conjugacy classes and each even
  weight ``k`` we can associate a corresponding rational period
  function. I.e. a rational function ``q`` of weight ``k`` which satisfies:
  ``q | S == 0`` and ``q + q|U + ... + q|U^(n-1) == 0``,
  where ``S``, ``U`` are the corresponding group elements and
  ``|`` is the usual `slash-operator` of weight ``k``.

  The set of all rational period function is expected
  to be generated by such functions.

  EXAMPLES::

      sage: from sage.modular.modform_hecketriangle.hecke_triangle_groups import HeckeTriangleGroup
      sage: G = HeckeTriangleGroup(n=5)
      sage: S = G.S()
      sage: U = G.U()

      sage: def is_rpf(f, k=None) -> bool:
      ....:     if not f + S.slash(f, k=k) == 0:
      ....:         return False
      ....:     return sum((U^m).slash(f, k=k) for m in range(G.n())) == 0

      sage: z = PolynomialRing(G.base_ring(), 'z').gen()
      sage: [is_rpf(1 - z^(-k), k=k) for k in range(-6, 6, 2)]  # long time
      [True, True, True, True, True, True]
      sage: [is_rpf(1/z, k=k) for k in range(-6, 6, 2)]
      [False, False, False, False, True, False]

      sage: el = G.V(2)
      sage: el.is_hecke_symmetric()
      False
      sage: rpf = el.rational_period_function(-4)
      sage: is_rpf(rpf)
      True
      sage: rpf
      -lam*z^4 + lam
      sage: rpf = el.rational_period_function(-2)
      sage: is_rpf(rpf)
      True
      sage: rpf
      (lam + 1)*z^2 - lam - 1
      sage: el.rational_period_function(0) == 0
      True
      sage: rpf = el.rational_period_function(2)
      sage: is_rpf(rpf)
      True
      sage: rpf
      ((lam + 1)*z^2 - lam - 1)/(lam*z^4 + (-lam - 2)*z^2 + lam)

      sage: el = G.V(2)*G.V(3)
      sage: el.is_hecke_symmetric()
      True
      sage: el.rational_period_function(-4) == 0
      True
      sage: rpf = el.rational_period_function(-2)
      sage: rpf
      (8*lam + 4)*z^2 - 8*lam - 4
      sage: rpf = el.rational_period_function(2)
      sage: is_rpf(rpf)
      True
      sage: rpf.denominator()
      (144*lam + 89)*z^8 + (-618*lam - 382)*z^6 + (951*lam + 588)*z^4 + (-618*lam - 382)*z^2 + 144*lam + 89
      sage: el.rational_period_function(4) == 0
      True

      sage: G = HeckeTriangleGroup(n=4)
      sage: G.rational_period_functions(k=4, D=12)
      [(z^4 - 1)/z^4]
      sage: G.rational_period_functions(k=2, D=14)
      [(z^2 - 1)/z^2, 1/z, (24*z^6 - 120*z^4 + 120*z^2 - 24)/(9*z^8 - 80*z^6 + 146*z^4 - 80*z^2 + 9), (24*z^6 - 120*z^4 + 120*z^2 - 24)/(9*z^8 - 80*z^6 + 146*z^4 - 80*z^2 + 9)]


- **Block decomposition of elements:**
  For each group element a very specific conjugacy representative
  can be obtained. For hyperbolic and parabolic elements the
  representative is a product ``V(j)``-matrices. They all
  have nonnegative trace and the number of factors is called
  the block length of the element (which is implemented).

  Note: For this decomposition special care is given to the
  sign (of the trace) of the matrices.

  The case ``n=infinity`` for everything above is not properly implemented.

  EXAMPLES::

      sage: from sage.modular.modform_hecketriangle.hecke_triangle_groups import HeckeTriangleGroup
      sage: G = HeckeTriangleGroup(n=7)
      sage: G.element_repr_method("block")

      sage: A = -G.V(2)*G.V(6)^3*G.V(3)
      sage: A
      -(T*S*T) * (V(6)^3*V(3)*V(2)) * (T*S*T)^(-1)
      sage: A.sign()
      -1
      sage: (L, R, sgn) = A.block_decomposition()
      sage: L
      ((-S*T^(-1)*S) * (V(6)^3) * (-S*T^(-1)*S)^(-1), (T*S*T*S*T) * (V(3)) * (T*S*T*S*T)^(-1), (T*S*T) * (V(2)) * (T*S*T)^(-1))
      sage: prod(L).sign()
      1
      sage: A == sgn * (R.acton(prod(L)))
      True
      sage: t = A.block_length()
      sage: t
      5
      sage: AA(A.discriminant()) >= AA(t^2 * G.lam() - 4)
      True


- **Class number and class representatives**:
  The block length provides a lower bound for the
  discriminant. This allows to enlist all (representatives of)
  matrices of (or up to) a given discriminant.

  Using the 1-1 correspondence with hyperbolic fixed points
  (and certain hyperbolic binary quadratic forms) this
  makes it possible to calculate the corresponding class number
  (number of conjugacy classes for a given discriminant).

  It also allows to list all occurring discriminants up
  to some bound. Or to enlist all reduced/simple elements
  resp. their corresponding hyperbolic fixed points
  for the given discriminant.

  Warning: The currently used algorithm is very slow!

  EXAMPLES::

      sage: from sage.modular.modform_hecketriangle.hecke_triangle_groups import HeckeTriangleGroup
      sage: G = HeckeTriangleGroup(n=4)
      sage: G.element_repr_method("basic")
      sage: G.is_discriminant(68)
      True
      sage: G.class_number(14)
      2
      sage: G.list_discriminants(D=68)
      [4, 12, 14, 28, 32, 46, 60, 68]
      sage: G.list_discriminants(D=0, hyperbolic=False, primitive=False)
      [-4, -2, 0]
      sage: G.class_number(68)
      4
      sage: sorted(G.class_representatives(68))
      [S*T^(-5)*S*T^(-1)*S, S*T^(-2)*S*T^(-1)*S*T, T*S*T^5, -S*T^(-1)*S*T^2*S*T]
      sage: R = G.reduced_elements(68)
      sage: all(v.is_reduced() for v in R)  # long time
      True
      sage: R = G.simple_elements(68)
      sage: all(v.is_simple() for v in R)  # long time
      True
      sage: G.element_repr_method("default")

      sage: G = HeckeTriangleGroup(n=5)
      sage: G.element_repr_method("basic")
      sage: G.list_discriminants(9*G.lam() + 5)
      [4*lam, 7*lam + 6, 9*lam + 5]
      sage: G.list_discriminants(D=0, hyperbolic=False, primitive=False)
      [-4, -lam - 2, lam - 3, 0]
      sage: G.class_number(9*G.lam() + 5)
      2
      sage: sorted(G.class_representatives(9*G.lam() + 5))
      [S*T^(-2)*S*T^(-1)*S, T*S*T^2]
      sage: R = G.reduced_elements(9*G.lam() + 5)
      sage: all(v.is_reduced() for v in R)  # long time
      True
      sage: R = G.simple_elements(7*G.lam() + 6)
      sage: for v in R: print(v.string_repr("default"))
      [lam + 2     lam]
      [    lam       1]
      [      1     lam]
      [    lam lam + 2]
      sage: G.element_repr_method("default")


Modular forms ring and spaces for Hecke triangle groups:
--------------------------------------------------------

- **Analytic type:**
  The analytic type of forms, including the behavior at infinity:

  - Meromorphic (and meromorphic at infinity)
  - Weakly holomorphic (holomorphic and meromorphic at infinity)
  - Holomorphic (and holomorphic at infinity)
  - Cuspidal (holomorphic and zero at infinity)

  Additionally the type specifies whether the form is modular or only quasi modular.

  EXAMPLES::

      sage: from sage.modular.modform_hecketriangle.analytic_type import AnalyticType
      sage: AnalyticType()(["quasi", "cusp"])
      quasi cuspidal


- **Modular form (for Hecke triangle groups):**
  A function of some analytic type which transforms like a modular form
  for the given group, weight ``k`` and multiplier ``epsilon``:

  - ``f(z+lambda) = f(lambda)``
  - ``f(-1/z) = epsilon * (z/i)^k * f(z)``

  The multiplier is either ``1`` or ``-1``.
  The weight is a rational number of the form ``4*(n*l+l')/(n-2) + (1-epsilon)*n/(n-2)``.
  If ``n`` is odd, then the multiplier is unique and given by ``(-1)^(k*(n-2)/2)``.
  The space of modular forms for a given group, weight and multiplier forms a module
  over the base ring. It is finite dimensional if the analytic type is ``holomorphic``.

  Modular forms can be constructed in several ways:

  - Using some already available construction function for modular forms
    (those function are available for all spaces/rings and in general
    do not return elements of the same parent)
  - Specifying the form as a rational function in the basic generators (see below)
  - For weakly holomorphic modular forms it is possible to exactly determine the
    form by specifying (sufficiently many) initial coefficients of its Fourier expansion.
  - There is even hope (no guarantee) to determine a (exact) form from
    the initial numerical coefficients (see below).
  - By specifying the coefficients with respect to a basis of the space
    (if the corresponding space supports coordinate vectors)
  - Arithmetic combination of forms or differential operators applied to forms

  The implementation is based on the implementation of the graded ring (see below).
  All calculations are exact (no precision argument is required).
  The analytic type of forms is checked during construction.
  The analytic type of parent spaces after arithmetic/differential operations
  with elements is changed (extended/reduced) accordingly.

  In particular it is possible to multiply arbitrary modular forms (and end up
  with an element of a modular forms space). If two forms of different
  weight/multiplier are added then an element of the corresponding
  modular forms ring is returned instead.

  Elements of modular forms spaces are represented by their Fourier expansion.

  EXAMPLES::

      sage: from sage.modular.modform_hecketriangle.space import CuspForms, ModularForms, MeromorphicModularForms
      sage: MeromorphicModularForms(n=4, k=8, ep=1)
      MeromorphicModularForms(n=4, k=8, ep=1) over Integer Ring
      sage: CF = CuspForms(n=7, k=12, ep=1)
      sage: CF
      CuspForms(n=7, k=12, ep=1) over Integer Ring

      sage: MF = ModularForms(k=12, ep=1)
      sage: (x,y,z,d) = MF.pol_ring().gens()

  Using existing functions::

      sage: CF.Delta()
      q + 17/(56*d)*q^2 + 88887/(2458624*d^2)*q^3 + 941331/(481890304*d^3)*q^4 + O(q^5)

  Using rational function in the basic generators::

      sage: MF(x^3)
      1 + 720*q + 179280*q^2 + 16954560*q^3 + 396974160*q^4 + O(q^5)

  Using Fourier expansions::

      sage: qexp = CF.Delta().q_expansion(prec=2)
      sage: qexp
      q + O(q^2)
      sage: qexp.parent()
      Power Series Ring in q over Fraction Field of Univariate Polynomial Ring in d over Integer Ring
      sage: MF(qexp)
      q - 24*q^2 + 252*q^3 - 1472*q^4 + O(q^5)

  Using coordinate vectors::

      sage: MF([0,1]) == MF.f_inf()
      True

  Using arithmetic expressions::

      sage: d = CF.get_d()
      sage: CF.f_rho()^7 / (d*CF.f_rho()^7 - d*CF.f_i()^2) == CF.j_inv()
      True
      sage: MF.E4().serre_derivative() == -1/3 * MF.E6()
      True


- **Hauptmodul:**
  The ``j-function`` for Hecke triangle groups is given by the unique Riemann map
  from the hyperbolic triangle with vertices at ``rho``, ``i`` and ``infinity`` to the
  upper half plane, normalized such that its Fourier coefficients are real and such
  that the first nontrivial Fourier coefficient is 1. The function extends to a
  completely invariant weakly holomorphic function from the upper half plane to the
  complex numbers. Another used normalization (in capital letters) is ``J(i)=1``.
  The coefficients of ``j`` are rational numbers up to a power of ``d=1/j(i)``
  which is only rational in the arithmetic cases ``n=3, 4, 6, infinity``.

  All Fourier coefficients of modular forms are based on the coefficients of ``j``.
  The coefficients of ``j`` are calculated by inverting the Fourier series of its
  inverse (the series inversion is also by far the most expensive operation of all).

  EXAMPLES::

      sage: from sage.modular.modform_hecketriangle.graded_ring import WeakModularFormsRing
      sage: from sage.modular.modform_hecketriangle.space import WeakModularForms
      sage: WeakModularForms(n=3, k=0, ep=1).j_inv()
      q^-1 + 744 + 196884*q + 21493760*q^2 + 864299970*q^3 + 20245856256*q^4 + O(q^5)
      sage: WeakModularFormsRing(n=7).j_inv()
      f_rho^7/(f_rho^7*d - f_i^2*d)
      sage: WeakModularFormsRing(n=7, red_hom=True).j_inv()
      q^-1 + 151/(392*d) + 165229/(2458624*d^2)*q + 107365/(15059072*d^3)*q^2 + 25493858865/(48358655787008*d^4)*q^3 + 2771867459/(92561489592320*d^5)*q^4 + O(q^5)


- **Basic generators:**
  There exist unique modular forms ``f_rho``, ``f_i`` and ``f_inf`` such that
  each has a simple zero at ``rho=exp(pi/n)``, ``i`` and ``infinity`` resp. and
  no other zeros. The forms are normalized such that their first Fourier coefficient
  is ``1``. They have the weight and multiplier ``(4/(n-2), 1)``, ``(2*n/(n-2), -1)``,
  ``(4*n/(n-2), 1)`` resp. and can be defined in terms of the Hauptmodul ``j``.

  EXAMPLES::

      sage: from sage.modular.modform_hecketriangle.graded_ring import ModularFormsRing
      sage: ModularFormsRing(n=5, red_hom=True).f_rho()
      1 + 7/(100*d)*q + 21/(160000*d^2)*q^2 + 1043/(192000000*d^3)*q^3 + 45479/(1228800000000*d^4)*q^4 + O(q^5)
      sage: ModularFormsRing(n=5, red_hom=True).f_i()
      1 - 13/(40*d)*q - 351/(64000*d^2)*q^2 - 13819/(76800000*d^3)*q^3 - 1163669/(491520000000*d^4)*q^4 + O(q^5)
      sage: ModularFormsRing(n=5, red_hom=True).f_inf()
      q - 9/(200*d)*q^2 + 279/(640000*d^2)*q^3 + 961/(192000000*d^3)*q^4 + O(q^5)
      sage: ModularFormsRing(n=5).f_inf()
      f_rho^5*d - f_i^2*d


- **Eisenstein series and Delta:**
  The Eisenstein series of weight ``2``, ``4`` and ``6`` exist for all ``n`` and
  are all implemented . Note that except for ``n=3`` the series ``E4`` and ``E6``
  do not coincide with ``f_rho`` and ``f_i``.

  Similarly there always exists a (generalization of) ``Delta``. Except for ``n=3``
  it also does not coincide with ``f_inf``.

  In general Eisenstein series of all even weights exist for all ``n``.
  In the non-arithmetic cases they are however very hard to determine
  (it's an open problem(?) and consequently not yet implemented,
  except for trivial one-dimensional cases).

  The Eisenstein series in the arithmetic cases ``n = 3, 4, 6`` are fully
  implemented though. Note that this requires a lot more work/effort
  for ``k != 2, 4, 6`` resp. for multidimensional spaces.

  The case ``n=infinity`` is a special case (since there are two cusps)
  and is not implemented yet.

  EXAMPLES::

      sage: from sage.modular.modform_hecketriangle.graded_ring import ModularFormsRing
      sage: from sage.modular.modform_hecketriangle.space import ModularForms
      sage: ModularFormsRing(n=5).E4()
      f_rho^3
      sage: ModularFormsRing(n=5).E6()
      f_rho^2*f_i
      sage: ModularFormsRing(n=5).Delta()
      f_rho^9*d - f_rho^4*f_i^2*d
      sage: ModularFormsRing(n=5).Delta() == ModularFormsRing(n=5).f_inf()*ModularFormsRing(n=5).f_rho()^4
      True

  The basic generators in some arithmetic cases::

      sage: ModularForms(n=3, k=6).E6()
      1 - 504*q - 16632*q^2 - 122976*q^3 - 532728*q^4 + O(q^5)
      sage: ModularForms(n=4, k=6).E6()
      1 - 56*q - 2296*q^2 - 13664*q^3 - 73976*q^4 + O(q^5)
      sage: ModularForms(n=infinity, k=4).E4()
      1 + 16*q + 112*q^2 + 448*q^3 + 1136*q^4 + O(q^5)

  General Eisenstein series in some arithmetic cases::

      sage: ModularFormsRing(n=4).EisensteinSeries(k=8) * 34
      25*f_rho^4 + 9*f_i^2
      sage: ModularForms(n=3, k=12).EisensteinSeries()
      1 + 65520/691*q + 134250480/691*q^2 + 11606736960/691*q^3 + 274945048560/691*q^4 + O(q^5)
      sage: ModularForms(n=6, k=12).EisensteinSeries()
      1 + 6552/50443*q + 13425048/50443*q^2 + 1165450104/50443*q^3 + 27494504856/50443*q^4 + O(q^5)
      sage: ModularForms(n=4, k=22, ep=-1).EisensteinSeries()
      1 - 184/53057489*q - 386252984/53057489*q^2 - 1924704989536/53057489*q^3 - 810031218278584/53057489*q^4 + O(q^5)


- **Generator for ``k=0``, ``ep=-1``:**
  If ``n`` is even then the space of weakly holomorphic modular forms of weight
  ``0`` and multiplier ``-1`` is not empty and generated by one element,
  denoted by ``g_inv``.

  EXAMPLES::

      sage: from sage.modular.modform_hecketriangle.space import WeakModularForms
      sage: WeakModularForms(n=4, k=0, ep=-1).g_inv()
      q^-1 - 24 - 3820*q - 100352*q^2 - 1217598*q^3 - 10797056*q^4 + O(q^5)
      sage: WeakModularFormsRing(n=8).g_inv()
      f_rho^4*f_i/(f_rho^8*d - f_i^2*d)


- **Quasi modular form (for Hecke triangle groups):**
  ``E2`` no longer transforms like a modular form but like a quasi modular form.
  More generally quasi modular forms are given in terms of modular forms and powers
  of ``E2``. E.g. a holomorphic quasi modular form is a sum of holomorphic modular
  forms multiplied with a power of ``E2`` such that the weights and multipliers match up.
  The space of quasi modular forms for a given group, weight and multiplier forms a
  module over the base ring. It is finite dimensional if the analytic type is
  ``holomorphic``.

  The implementation and construction are analogous to modular forms (see above).
  In particular construction of quasi weakly holomorphic forms by their initial
  Laurent coefficients is supported as well!

  EXAMPLES::

      sage: from sage.modular.modform_hecketriangle.graded_ring import ModularFormsRing
      sage: from sage.modular.modform_hecketriangle.space import QuasiCuspForms, QuasiModularForms, QuasiWeakModularForms
      sage: QuasiCuspForms(n=7, k=12, ep=1)
      QuasiCuspForms(n=7, k=12, ep=1) over Integer Ring
      sage: QuasiModularForms(n=4, k=8, ep=-1)
      QuasiModularForms(n=4, k=8, ep=-1) over Integer Ring

      sage: QuasiModularForms(n=4, k=2, ep=-1).E2()
      1 - 8*q - 40*q^2 - 32*q^3 - 104*q^4 + O(q^5)

  A quasi weak form can be constructed by using its initial Laurent expansion::

      sage: QF = QuasiWeakModularForms(n=8, k=10/3, ep=-1)
      sage: qexp = (QF.quasi_part_gens(min_exp=-1)[4]).q_expansion(prec=5)
      sage: qexp
      q^-1 - 19/(64*d) - 7497/(262144*d^2)*q + 15889/(8388608*d^3)*q^2 + 543834047/(1649267441664*d^4)*q^3 + 711869853/(43980465111040*d^5)*q^4 + O(q^5)
      sage: qexp.parent()
      Laurent Series Ring in q over Fraction Field of Univariate Polynomial Ring in d over Integer Ring
      sage: QF(qexp).as_ring_element()
      f_rho^3*f_i*E2^2/(f_rho^8*d - f_i^2*d)
      sage: QF(qexp).reduced_parent()
      QuasiWeakModularForms(n=8, k=10/3, ep=-1) over Integer Ring

  Derivatives of (quasi weak) modular forms are again quasi (weak) modular forms::

      sage: CF.f_inf().derivative() == CF.f_inf()*CF.E2()
      True


- **Ring of (quasi) modular forms:**
  The ring of (quasi) modular forms for a given analytic type and Hecke triangle group.
  In fact it is a graded algebra over the base ring where the grading is over
  ``1/(n-2)*Z x Z/(2Z)`` corresponding to the weight and multiplier.
  A ring element is thus a finite linear combination of (quasi) modular forms
  of (possibly) varying weights and multipliers.

  Each ring element is represented as a rational function in the
  generators ``f_rho``, ``f_i`` and ``E2``. The representations and arithmetic
  operations are exact (no precision argument is required).

  Elements of the ring are represented by the rational function in the generators.

  If the parameter ``red_hom`` is set to ``True`` (default: ``False``) then
  operations with homogeneous elements try to return an element of the corresponding
  vector space (if the element is homogeneous) instead of the forms ring.
  It is also easier to use the forms ring with ``red_hom=True`` to construct known
  forms (since then it is not required to specify the weight and multiplier).

  EXAMPLES::

      sage: from sage.modular.modform_hecketriangle.graded_ring import QuasiModularFormsRing, ModularFormsRing
      sage: QuasiModularFormsRing(n=5, red_hom=True)
      QuasiModularFormsRing(n=5) over Integer Ring
      sage: ModularFormsRing()
      ModularFormsRing(n=3) over Integer Ring
      sage: (x,y,z,d) = ModularFormsRing().pol_ring().gens()

      sage: ModularFormsRing()(x+y)
      f_rho + f_i

      sage: QuasiModularFormsRing(n=5, red_hom=True)(x^5-y^2).reduce()
      1/d*q - 9/(200*d^2)*q^2 + 279/(640000*d^3)*q^3 + 961/(192000000*d^4)*q^4 + O(q^5)


- **Construction of modular forms spaces and rings:**
  There are functorial constructions behind all forms spaces and rings
  which assure that arithmetic operations between those spaces and rings
  work and fit into the coercion framework. In particular ring elements
  are interpreted as constant modular forms in this context and base
  extensions are done if necessary.


- **Fourier expansion of (quasi) modular forms (for Hecke triangle groups):**
  Each (quasi) modular form (in fact each ring element) possesses a Fourier
  expansion of the form ``sum_{n>=n_0} a_n q^n``, where ``n_0`` is an integer,
  ``q=exp(2*pi*i*z/lambda)`` and the coefficients ``a_n`` are rational numbers
  (or more generally an extension of rational numbers) up to a power of ``d``,
  where ``d`` is the (possibly) transcendental parameter described above.
  I.e. the coefficient ring is given by ``Frac(R)(d)``.

  The coefficients are calculated exactly in terms of the (formal) parameter
  ``d``. The expansion is calculated exactly up to the specified precision.
  It is also possible to get a Fourier expansion where ``d`` is evaluated
  to its numerical approximation.

  EXAMPLES::

      sage: from sage.modular.modform_hecketriangle.graded_ring import ModularFormsRing, QuasiModularFormsRing
      sage: ModularFormsRing(n=4).j_inv().q_expansion(prec=3)
      q^-1 + 13/(32*d) + 1093/(16384*d^2)*q + 47/(8192*d^3)*q^2 + O(q^3)
      sage: QuasiModularFormsRing(n=5).E2().q_expansion(prec=3)
      1 - 9/(200*d)*q - 369/(320000*d^2)*q^2 + O(q^3)
      sage: QuasiModularFormsRing(n=5).E2().q_expansion_fixed_d(prec=3)
      1.000000000000... - 6.380956565426...*q - 23.18584547617...*q^2 + O(q^3)


- **Evaluation of forms:**
  (Quasi) modular forms (and also ring elements) can be viewed as
  functions from the upper half plane and can be numerically evaluated
  by using the Fourier expansion.

  The evaluation uses the (quasi) modularity properties (if possible)
  for a faster and more precise evaluation. The precision of the result
  depends both on the numerical precision and on the default precision
  used for the Fourier expansion.

  EXAMPLES::

      sage: from sage.modular.modform_hecketriangle.graded_ring import ModularFormsRing
      sage: f_i = ModularFormsRing(n=4).f_i()
      sage: f_i(i)
      0
      sage: f_i(infinity)
      1
      sage: f_i(1/7 + 0.01*i)
      32189.02016723... + 21226.62951394...*I


- **L-functions of forms:**
  Using the (pari based) function ``Dokchitser`` L-functions of non-constant
  holomorphic modular forms are supported for all values of ``n``.

  Note: For non-arithmetic groups this involves an irrational conductor.
  The conductor for the arithmetic groups ``n = 3, 4, 6, infinity`` is
  ``1, 2, 3, 4`` respectively.

  EXAMPLES::

      sage: from sage.modular.modform.eis_series import eisenstein_series_lseries
      sage: from sage.modular.modform_hecketriangle.space import ModularForms
      sage: f = ModularForms(n=3, k=4).E4()/240
      sage: L = f.lseries()
      sage: L.conductor
      1
      sage: L.check_functional_equation() < 2^(-50)
      True
      sage: L(1)
      -0.0304484570583...
      sage: abs(L(1) - eisenstein_series_lseries(4)(1)) < 2^(-53)
      True
      sage: L.taylor_series(1, 3)
      -0.0304484570583... - 0.0504570844798...*z - 0.0350657360354...*z^2 + O(z^3)
      sage: coeffs = f.q_expansion_vector(min_exp=0, max_exp=20, fix_d=True)
      sage: abs(L(10) - sum([coeffs[k] * ZZ(k)^(-10) for k in range(1,len(coeffs))]).n(53)) < 10^(-7)
      True

      sage: L = ModularForms(n=6, k=6, ep=-1).E6().lseries(num_prec=200)
      sage: L.conductor
      3
      sage: L.check_functional_equation() < 2^(-180)
      True
      sage: L.eps
      -1
      sage: abs(L(3)) < 2^(-180)
      True

      sage: L = ModularForms(n=17, k=12).Delta().lseries()
      sage: L.conductor
      3.86494445880...
      sage: L.check_functional_equation() < 2^(-50)
      True
      sage: L.taylor_series(6, 3)
      2.15697985314... - 1.17385918996...*z + 0.605865993050...*z^2 + O(z^3)

      sage: L = ModularForms(n=infinity, k=2, ep=-1).f_i().lseries()
      sage: L.conductor
      4
      sage: L.check_functional_equation() < 2^(-50)
      True
      sage: L.taylor_series(1, 3)
      0.000000000000... + 5.76543616701...*z + 9.92776715593...*z^2 + O(z^3)


- **(Serre) derivatives:**
  Derivatives and Serre derivatives of forms can be calculated.
  The analytic type is extended accordingly.

  EXAMPLES::

      sage: from sage.modular.modform_hecketriangle.graded_ring import ModularFormsRing
      sage: from sage.modular.modform_hecketriangle.space import QuasiModularForms
      sage: f_inf = ModularFormsRing(n=4, red_hom=True).f_inf()
      sage: f_inf.derivative()/f_inf == QuasiModularForms(n=4, k=2, ep=-1).E2()
      True
      sage: ModularFormsRing().E4().serre_derivative() == -1/3 * ModularFormsRing().E6()
      True


- **Basis for weakly holomorphic modular forms and Faber polynomials:**
  (Natural) generators of weakly holomorphic modular forms can
  be obtained using the corresponding generalized Faber polynomials.

  EXAMPLES::

      sage: from sage.modular.modform_hecketriangle.space import WeakModularForms, CuspForms
      sage: MF = WeakModularForms(n=5, k=62/3, ep=-1)
      sage: MF.disp_prec(MF._l1+2)

      sage: MF.F_basis(2)
      q^2 - 41/(200*d)*q^3 + O(q^4)
      sage: MF.F_basis(1)
      q - 13071/(640000*d^2)*q^3 + O(q^4)
      sage: MF.F_basis(-0)
      1 - 277043/(192000000*d^3)*q^3 + O(q^4)
      sage: MF.F_basis(-2)
      q^-2 - 162727620113/(40960000000000000*d^5)*q^3 + O(q^4)


- **Basis for quasi weakly holomorphic modular forms:**
  (Natural) generators of quasi weakly holomorphic modular forms can
  also be obtained. In most cases it is even possible to find a basis consisting
  of elements with only one non-trivial Laurent coefficient (up to some coefficient).

  EXAMPLES::

      sage: from sage.modular.modform_hecketriangle.space import QuasiWeakModularForms
      sage: QF = QuasiWeakModularForms(n=8, k=10/3, ep=-1)
      sage: QF.default_prec(1)
      sage: QF.quasi_part_gens(min_exp=-1)
      (q^-1 + O(q),
       1 + O(q),
       q^-1 - 9/(128*d) + O(q),
       1 + O(q),
       q^-1 - 19/(64*d) + O(q),
       q^-1 + 1/(64*d) + O(q))
      sage: QF.default_prec(QF.required_laurent_prec(min_exp=-1))
      sage: QF.q_basis(min_exp=-1)    # long time
      [q^-1 + O(q^5),
       1 + O(q^5),
       q + O(q^5),
       q^2 + O(q^5),
       q^3 + O(q^5),
       q^4 + O(q^5)]


- **Dimension and basis for holomorphic or cuspidal (quasi) modular forms:**
  For finite dimensional spaces the dimension and a basis can be obtained.

  EXAMPLES::

      sage: from sage.modular.modform_hecketriangle.space import QuasiModularForms
      sage: MF = QuasiModularForms(n=5, k=6, ep=-1)
      sage: MF.dimension()
      3
      sage: MF.default_prec(2)
      sage: MF.gens()
      (1 - 37/(200*d)*q + O(q^2),
       1 + 33/(200*d)*q + O(q^2),
       1 - 27/(200*d)*q + O(q^2))


- **Coordinate vectors for (quasi) holomorphic modular forms and (quasi) cusp forms:**
  For (quasi) holomorphic modular forms and (quasi) cusp forms it is possible
  to determine the coordinate vectors of elements with respect to the basis.

  EXAMPLES::

      sage: from sage.modular.modform_hecketriangle.space import ModularForms
      sage: ModularForms(n=7, k=12, ep=1).dimension()
      3
      sage: ModularForms(n=7, k=12, ep=1).Delta().coordinate_vector()
      (0, 1, 17/(56*d))

      sage: from sage.modular.modform_hecketriangle.space import QuasiCuspForms
      sage: MF = QuasiCuspForms(n=7, k=20, ep=1)
      sage: MF.dimension()
      13
      sage: el = MF(MF.Delta()*MF.E2()^4 + MF.Delta()*MF.E2()*MF.E6())
      sage: el.coordinate_vector()    # long time
      (0, 0, 0, 1, 29/(196*d), 0, 0, 0, 0, 1, 17/(56*d), 0, 0)


- **Subspaces:**
  It is possible to construct subspaces of (quasi) holomorphic modular forms
  or (quasi) cusp forms spaces with respect to a specified basis of the
  corresponding ambient space. The subspaces also support coordinate
  vectors with respect to its basis.

  EXAMPLES::

      sage: from sage.modular.modform_hecketriangle.space import ModularForms
      sage: MF = ModularForms(n=7, k=12, ep=1)
      sage: subspace = MF.subspace([MF.E4()^3, MF.Delta()])
      sage: subspace
      Subspace of dimension 2 of ModularForms(n=7, k=12, ep=1) over Integer Ring
      sage: el = subspace(MF.E6()^2)
      sage: el.coordinate_vector()
      (1, -61/(196*d))
      sage: el.ambient_coordinate_vector()
      (1, -61/(196*d), -51187/(614656*d^2))

      sage: from sage.modular.modform_hecketriangle.space import QuasiCuspForms
      sage: MF = QuasiCuspForms(n=7, k=20, ep=1)
      sage: subspace = MF.subspace([MF.Delta()*MF.E2()^2*MF.E4(), MF.Delta()*MF.E2()^4])    # long time
      sage: subspace    # long time
      Subspace of dimension 2 of QuasiCuspForms(n=7, k=20, ep=1) over Integer Ring
      sage: el = subspace(MF.Delta()*MF.E2()^4)    # long time
      sage: el.coordinate_vector()    # long time
      (0, 1)
      sage: el.ambient_coordinate_vector()    # long time
      (0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 17/(56*d), 0, 0)


- **Theta subgroup:**
  The Hecke triangle group corresponding to ``n=infinity`` is also
  completely supported. In particular the (special) behavior around
  the cusp ``-1`` is considered and can be specified.

  EXAMPLES::

      sage: from sage.modular.modform_hecketriangle.graded_ring import QuasiMeromorphicModularFormsRing
      sage: MR = QuasiMeromorphicModularFormsRing(n=infinity, red_hom=True)
      sage: MR
      QuasiMeromorphicModularFormsRing(n=+Infinity) over Integer Ring
      sage: j_inv = MR.j_inv().full_reduce()
      sage: f_i = MR.f_i().full_reduce()
      sage: E4 = MR.E4().full_reduce()
      sage: E2 = MR.E2().full_reduce()

      sage: j_inv
      q^-1 + 24 + 276*q + 2048*q^2 + 11202*q^3 + 49152*q^4 + O(q^5)
      sage: MR.f_rho() == MR(1)
      True
      sage: E4
      1 + 16*q + 112*q^2 + 448*q^3 + 1136*q^4 + O(q^5)
      sage: f_i
      1 - 24*q + 24*q^2 - 96*q^3 + 24*q^4 + O(q^5)
      sage: E2
      1 - 8*q - 8*q^2 - 32*q^3 - 40*q^4 + O(q^5)
      sage: E4.derivative() == E4 * (E2 - f_i)
      True
      sage: f_i.serre_derivative() == -1/2 * E4
      True
      sage: MF = f_i.serre_derivative().parent()
      sage: MF
      ModularForms(n=+Infinity, k=4, ep=1) over Integer Ring
      sage: MF.dimension()
      2
      sage: MF.gens()
      (1 + 240*q^2 + 2160*q^4 + O(q^5), q - 8*q^2 + 28*q^3 - 64*q^4 + O(q^5))
      sage: E4(i)
      1.941017189...
      sage: E4.order_at(-1)
      1

      sage: MF = (E2/E4).reduced_parent()
      sage: MF.quasi_part_gens(order_1=-1)
      (1 - 40*q + 552*q^2 - 4896*q^3 + 33320*q^4 + O(q^5),
       1 - 24*q + 264*q^2 - 2016*q^3 + 12264*q^4 + O(q^5))
      sage: prec = MF.required_laurent_prec(order_1=-1)
      sage: qexp = (E2/E4).q_expansion(prec=prec)
      sage: qexp
      1 - 3/(8*d)*q + O(q^2)
      sage: MF.construct_quasi_form(qexp, order_1=-1) == E2/E4
      True
      sage: MF.disp_prec(6)
      sage: MF.q_basis(m=-1, order_1=-1, min_exp=-1)
      q^-1 - 203528/7*q^5 + O(q^6)

  Elements with respect to the full group are automatically coerced
  to elements of the Theta subgroup if necessary::

      sage: el = QuasiMeromorphicModularFormsRing(n=3).Delta().full_reduce() + E2
      sage: el
      (E4*f_i^4 - 2*E4^2*f_i^2 + E4^3 + 4096*E2)/4096
      sage: el.parent()
      QuasiModularFormsRing(n=+Infinity) over Integer Ring


- **Determine exact coefficients from numerical ones:**
  There is some experimental support for replacing numerical coefficients with
  corresponding exact coefficients. There is however NO guarantee that
  the procedure will work (and most probably there are cases where it won't).

  EXAMPLES::

     sage: from sage.modular.modform_hecketriangle.space import WeakModularForms, QuasiCuspForms
     sage: WF = WeakModularForms(n=14)
     sage: qexp = WF.J_inv().q_expansion_fixed_d(d_num_prec=1000)
     sage: qexp.parent()
     Laurent Series Ring in q over Real Field with 1000 bits of precision
     sage: qexp_int = WF.rationalize_series(qexp)
     doctest:...: UserWarning: Using an experimental rationalization of coefficients, please check the result for correctness!
     sage: qexp_int.parent()
     Laurent Series Ring in q over Fraction Field of Univariate Polynomial Ring in d over Integer Ring
     sage: qexp_int == WF.J_inv().q_expansion()
     True
     sage: WF(qexp_int) == WF.J_inv()
     True

     sage: QF = QuasiCuspForms(n=8, k=22/3, ep=-1)
     sage: el = QF(QF.f_inf()*QF.E2())
     sage: qexp = el.q_expansion_fixed_d(d_num_prec=1000)
     sage: qexp_int = QF.rationalize_series(qexp)
     sage: qexp_int == el.q_expansion()
     True
     sage: QF(qexp_int) == el
     True


Future ideas:
-------------

- Complete support for the case ``n=infinity`` (e.g. lambda-CF)

- Properly implemented lambda-CF

- Binary quadratic forms for Hecke triangle groups

- Cycle integrals

- Maybe: Proper spaces (with coordinates) for (quasi) weakly holomorphic forms with bounds on the initial Fourier exponent

- Support for general triangle groups (hard)

- Support for "congruence" subgroups (hard)
"""



# distutils: libraries = flint
# distutils: depends = flint/arf.h

################################################################################
# This file is auto-generated by the script
#   SAGE_ROOT/tools/flint-autogen/flint_autogen.py.
# From the commit 3e2c3a3e091106a25ca9c6fba28e02f2cbcd654a
# Do not modify by hand! Fix and rerun the script instead.
################################################################################

from libc.stdio cimport FILE
from sage.libs.gmp.types cimport *
from sage.libs.mpfr.types cimport *
from sage.libs.flint.types cimport *

cdef extern from "flint_wrap.h":
    void arf_init(arf_t x) noexcept
    void arf_clear(arf_t x) noexcept
    slong arf_allocated_bytes(const arf_t x) noexcept
    void arf_zero(arf_t res) noexcept
    void arf_one(arf_t res) noexcept
    void arf_pos_inf(arf_t res) noexcept
    void arf_neg_inf(arf_t res) noexcept
    void arf_nan(arf_t res) noexcept
    bint arf_is_zero(const arf_t x) noexcept
    bint arf_is_one(const arf_t x) noexcept
    bint arf_is_pos_inf(const arf_t x) noexcept
    bint arf_is_neg_inf(const arf_t x) noexcept
    bint arf_is_nan(const arf_t x) noexcept
    bint arf_is_inf(const arf_t x) noexcept
    bint arf_is_normal(const arf_t x) noexcept
    bint arf_is_special(const arf_t x) noexcept
    bint arf_is_finite(const arf_t x) noexcept
    void arf_set(arf_t res, const arf_t x) noexcept
    void arf_set_mpz(arf_t res, const mpz_t x) noexcept
    void arf_set_fmpz(arf_t res, const fmpz_t x) noexcept
    void arf_set_ui(arf_t res, ulong x) noexcept
    void arf_set_si(arf_t res, slong x) noexcept
    void arf_set_mpfr(arf_t res, const mpfr_t x) noexcept
    void arf_set_d(arf_t res, double x) noexcept
    void arf_swap(arf_t x, arf_t y) noexcept
    void arf_init_set_ui(arf_t res, ulong x) noexcept
    void arf_init_set_si(arf_t res, slong x) noexcept
    int arf_set_round(arf_t res, const arf_t x, slong prec, arf_rnd_t rnd) noexcept
    int arf_set_round_si(arf_t res, slong x, slong prec, arf_rnd_t rnd) noexcept
    int arf_set_round_ui(arf_t res, ulong x, slong prec, arf_rnd_t rnd) noexcept
    int arf_set_round_mpz(arf_t res, const mpz_t x, slong prec, arf_rnd_t rnd) noexcept
    int arf_set_round_fmpz(arf_t res, const fmpz_t x, slong prec, arf_rnd_t rnd) noexcept
    void arf_set_si_2exp_si(arf_t res, slong m, slong e) noexcept
    void arf_set_ui_2exp_si(arf_t res, ulong m, slong e) noexcept
    void arf_set_fmpz_2exp(arf_t res, const fmpz_t m, const fmpz_t e) noexcept
    int arf_set_round_fmpz_2exp(arf_t res, const fmpz_t x, const fmpz_t e, slong prec, arf_rnd_t rnd) noexcept
    void arf_get_fmpz_2exp(fmpz_t m, fmpz_t e, const arf_t x) noexcept
    void arf_frexp(arf_t m, fmpz_t e, const arf_t x) noexcept
    double arf_get_d(const arf_t x, arf_rnd_t rnd) noexcept
    int arf_get_mpfr(mpfr_t res, const arf_t x, mpfr_rnd_t rnd) noexcept
    int arf_get_fmpz(fmpz_t res, const arf_t x, arf_rnd_t rnd) noexcept
    slong arf_get_si(const arf_t x, arf_rnd_t rnd) noexcept
    int arf_get_fmpz_fixed_fmpz(fmpz_t res, const arf_t x, const fmpz_t e) noexcept
    int arf_get_fmpz_fixed_si(fmpz_t res, const arf_t x, slong e) noexcept
    void arf_floor(arf_t res, const arf_t x) noexcept
    void arf_ceil(arf_t res, const arf_t x) noexcept
    void arf_get_fmpq(fmpq_t res, const arf_t x) noexcept
    bint arf_equal(const arf_t x, const arf_t y) noexcept
    bint arf_equal_si(const arf_t x, slong y) noexcept
    bint arf_equal_ui(const arf_t x, ulong y) noexcept
    bint arf_equal_d(const arf_t x, double y) noexcept
    int arf_cmp(const arf_t x, const arf_t y) noexcept
    int arf_cmp_si(const arf_t x, slong y) noexcept
    int arf_cmp_ui(const arf_t x, ulong y) noexcept
    int arf_cmp_d(const arf_t x, double y) noexcept
    int arf_cmpabs(const arf_t x, const arf_t y) noexcept
    int arf_cmpabs_ui(const arf_t x, ulong y) noexcept
    int arf_cmpabs_d(const arf_t x, double y) noexcept
    int arf_cmpabs_mag(const arf_t x, const mag_t y) noexcept
    int arf_cmp_2exp_si(const arf_t x, slong e) noexcept
    int arf_cmpabs_2exp_si(const arf_t x, slong e) noexcept
    int arf_sgn(const arf_t x) noexcept
    void arf_min(arf_t res, const arf_t a, const arf_t b) noexcept
    void arf_max(arf_t res, const arf_t a, const arf_t b) noexcept
    slong arf_bits(const arf_t x) noexcept
    bint arf_is_int(const arf_t x) noexcept
    bint arf_is_int_2exp_si(const arf_t x, slong e) noexcept
    void arf_abs_bound_lt_2exp_fmpz(fmpz_t res, const arf_t x) noexcept
    void arf_abs_bound_le_2exp_fmpz(fmpz_t res, const arf_t x) noexcept
    slong arf_abs_bound_lt_2exp_si(const arf_t x) noexcept
    void arf_get_mag(mag_t res, const arf_t x) noexcept
    void arf_get_mag_lower(mag_t res, const arf_t x) noexcept
    void arf_set_mag(arf_t res, const mag_t x) noexcept
    void mag_init_set_arf(mag_t res, const arf_t x) noexcept
    void mag_fast_init_set_arf(mag_t res, const arf_t x) noexcept
    void arf_mag_set_ulp(mag_t res, const arf_t x, slong prec) noexcept
    void arf_mag_add_ulp(mag_t res, const mag_t x, const arf_t y, slong prec) noexcept
    void arf_mag_fast_add_ulp(mag_t res, const mag_t x, const arf_t y, slong prec) noexcept
    void arf_init_set_shallow(arf_t z, const arf_t x) noexcept
    void arf_init_set_mag_shallow(arf_t z, const mag_t x) noexcept
    void arf_init_neg_shallow(arf_t z, const arf_t x) noexcept
    void arf_init_neg_mag_shallow(arf_t z, const mag_t x) noexcept
    void arf_randtest(arf_t res, flint_rand_t state, slong bits, slong mag_bits) noexcept
    void arf_randtest_not_zero(arf_t res, flint_rand_t state, slong bits, slong mag_bits) noexcept
    void arf_randtest_special(arf_t res, flint_rand_t state, slong bits, slong mag_bits) noexcept
    void arf_urandom(arf_t res, flint_rand_t state, slong bits, arf_rnd_t rnd) noexcept
    void arf_debug(const arf_t x) noexcept
    void arf_print(const arf_t x) noexcept
    void arf_printd(const arf_t x, slong d) noexcept
    char * arf_get_str(const arf_t x, slong d) noexcept
    void arf_fprint(FILE * file, const arf_t x) noexcept
    void arf_fprintd(FILE * file, const arf_t y, slong d) noexcept
    char * arf_dump_str(const arf_t x) noexcept
    int arf_load_str(arf_t x, const char * str) noexcept
    int arf_dump_file(FILE * stream, const arf_t x) noexcept
    int arf_load_file(arf_t x, FILE * stream) noexcept
    void arf_abs(arf_t res, const arf_t x) noexcept
    void arf_neg(arf_t res, const arf_t x) noexcept
    int arf_neg_round(arf_t res, const arf_t x, slong prec, arf_rnd_t rnd) noexcept
    int arf_add(arf_t res, const arf_t x, const arf_t y, slong prec, arf_rnd_t rnd) noexcept
    int arf_add_si(arf_t res, const arf_t x, slong y, slong prec, arf_rnd_t rnd) noexcept
    int arf_add_ui(arf_t res, const arf_t x, ulong y, slong prec, arf_rnd_t rnd) noexcept
    int arf_add_fmpz(arf_t res, const arf_t x, const fmpz_t y, slong prec, arf_rnd_t rnd) noexcept
    int arf_add_fmpz_2exp(arf_t res, const arf_t x, const fmpz_t y, const fmpz_t e, slong prec, arf_rnd_t rnd) noexcept
    int arf_sub(arf_t res, const arf_t x, const arf_t y, slong prec, arf_rnd_t rnd) noexcept
    int arf_sub_si(arf_t res, const arf_t x, slong y, slong prec, arf_rnd_t rnd) noexcept
    int arf_sub_ui(arf_t res, const arf_t x, ulong y, slong prec, arf_rnd_t rnd) noexcept
    int arf_sub_fmpz(arf_t res, const arf_t x, const fmpz_t y, slong prec, arf_rnd_t rnd) noexcept
    void arf_mul_2exp_si(arf_t res, const arf_t x, slong e) noexcept
    void arf_mul_2exp_fmpz(arf_t res, const arf_t x, const fmpz_t e) noexcept
    int arf_mul(arf_t res, const arf_t x, const arf_t y, slong prec, arf_rnd_t rnd) noexcept
    int arf_mul_ui(arf_t res, const arf_t x, ulong y, slong prec, arf_rnd_t rnd) noexcept
    int arf_mul_si(arf_t res, const arf_t x, slong y, slong prec, arf_rnd_t rnd) noexcept
    int arf_mul_mpz(arf_t res, const arf_t x, const mpz_t y, slong prec, arf_rnd_t rnd) noexcept
    int arf_mul_fmpz(arf_t res, const arf_t x, const fmpz_t y, slong prec, arf_rnd_t rnd) noexcept
    int arf_addmul(arf_t z, const arf_t x, const arf_t y, slong prec, arf_rnd_t rnd) noexcept
    int arf_addmul_ui(arf_t z, const arf_t x, ulong y, slong prec, arf_rnd_t rnd) noexcept
    int arf_addmul_si(arf_t z, const arf_t x, slong y, slong prec, arf_rnd_t rnd) noexcept
    int arf_addmul_mpz(arf_t z, const arf_t x, const mpz_t y, slong prec, arf_rnd_t rnd) noexcept
    int arf_addmul_fmpz(arf_t z, const arf_t x, const fmpz_t y, slong prec, arf_rnd_t rnd) noexcept
    int arf_submul(arf_t z, const arf_t x, const arf_t y, slong prec, arf_rnd_t rnd) noexcept
    int arf_submul_ui(arf_t z, const arf_t x, ulong y, slong prec, arf_rnd_t rnd) noexcept
    int arf_submul_si(arf_t z, const arf_t x, slong y, slong prec, arf_rnd_t rnd) noexcept
    int arf_submul_mpz(arf_t z, const arf_t x, const mpz_t y, slong prec, arf_rnd_t rnd) noexcept
    int arf_submul_fmpz(arf_t z, const arf_t x, const fmpz_t y, slong prec, arf_rnd_t rnd) noexcept
    int arf_fma(arf_t res, const arf_t x, const arf_t y, const arf_t z, slong prec, arf_rnd_t rnd) noexcept
    int arf_sosq(arf_t res, const arf_t x, const arf_t y, slong prec, arf_rnd_t rnd) noexcept
    int arf_sum(arf_t res, arf_srcptr terms, slong len, slong prec, arf_rnd_t rnd) noexcept
    void arf_approx_dot(arf_t res, const arf_t initial, int subtract, arf_srcptr x, slong xstep, arf_srcptr y, slong ystep, slong len, slong prec, arf_rnd_t rnd) noexcept
    int arf_div(arf_t res, const arf_t x, const arf_t y, slong prec, arf_rnd_t rnd) noexcept
    int arf_div_ui(arf_t res, const arf_t x, ulong y, slong prec, arf_rnd_t rnd) noexcept
    int arf_ui_div(arf_t res, ulong x, const arf_t y, slong prec, arf_rnd_t rnd) noexcept
    int arf_div_si(arf_t res, const arf_t x, slong y, slong prec, arf_rnd_t rnd) noexcept
    int arf_si_div(arf_t res, slong x, const arf_t y, slong prec, arf_rnd_t rnd) noexcept
    int arf_div_fmpz(arf_t res, const arf_t x, const fmpz_t y, slong prec, arf_rnd_t rnd) noexcept
    int arf_fmpz_div(arf_t res, const fmpz_t x, const arf_t y, slong prec, arf_rnd_t rnd) noexcept
    int arf_fmpz_div_fmpz(arf_t res, const fmpz_t x, const fmpz_t y, slong prec, arf_rnd_t rnd) noexcept
    int arf_sqrt(arf_t res, const arf_t x, slong prec, arf_rnd_t rnd) noexcept
    int arf_sqrt_ui(arf_t res, ulong x, slong prec, arf_rnd_t rnd) noexcept
    int arf_sqrt_fmpz(arf_t res, const fmpz_t x, slong prec, arf_rnd_t rnd) noexcept
    int arf_rsqrt(arf_t res, const arf_t x, slong prec, arf_rnd_t rnd) noexcept
    int arf_root(arf_t res, const arf_t x, ulong k, slong prec, arf_rnd_t rnd) noexcept
    int arf_complex_mul(arf_t e, arf_t f, const arf_t a, const arf_t b, const arf_t c, const arf_t d, slong prec, arf_rnd_t rnd) noexcept
    int arf_complex_mul_fallback(arf_t e, arf_t f, const arf_t a, const arf_t b, const arf_t c, const arf_t d, slong prec, arf_rnd_t rnd) noexcept
    int arf_complex_sqr(arf_t e, arf_t f, const arf_t a, const arf_t b, slong prec, arf_rnd_t rnd) noexcept
    int _arf_get_integer_mpn(mp_ptr y, mp_srcptr xp, mp_size_t xn, slong exp) noexcept
    int _arf_set_mpn_fixed(arf_t z, mp_srcptr xp, mp_size_t xn, mp_size_t fixn, int negative, slong prec, arf_rnd_t rnd) noexcept
    int _arf_set_round_ui(arf_t z, ulong x, int sgnbit, slong prec, arf_rnd_t rnd) noexcept
    int _arf_set_round_uiui(arf_t z, slong * fix, mp_limb_t hi, mp_limb_t lo, int sgnbit, slong prec, arf_rnd_t rnd) noexcept
    int _arf_set_round_mpn(arf_t z, slong * exp_shift, mp_srcptr x, mp_size_t xn, int sgnbit, slong prec, arf_rnd_t rnd) noexcept



# sage.doctest: needs sage.libs.flint sage.graphs
"""
Matching polynomial

This module contains the following methods:

.. csv-table::
    :class: contentstable
    :widths: 30, 70
    :delim: |

    :meth:`matching_polynomial` | Compute the matching polynomial of a given graph.
    :meth:`complete_poly` | Compute the matching polynomial of the complete graph on `n` vertices.

AUTHORS:

- Robert Miller, Tom Boothby - original implementation

REFERENCE:

[God1993]_


Methods
-------
"""

# ****************************************************************************
#       Copyright (C) 2010 Robert Miller
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from cysignals.memory cimport check_allocarray, sig_free
from cysignals.signals cimport sig_on, sig_off

from sage.rings.polynomial.polynomial_ring import polygen
from sage.rings.integer_ring import ZZ
from sage.rings.integer cimport Integer

from sage.libs.flint.fmpz cimport *
from sage.libs.flint.fmpz_poly cimport *
from sage.libs.flint.fmpz_poly_sage cimport *


x = polygen(ZZ, 'x')


def matching_polynomial(G, complement=True, name=None):
    r"""
    Compute the matching polynomial of the graph `G`.

    If `p(G, k)` denotes the number of `k`-matchings (matchings with `k` edges)
    in `G`, then the matching polynomial is defined as [God1993]_:

    .. MATH::

        \mu(x)=\sum_{k \geq 0} (-1)^k p(G,k) x^{n-2k}

    INPUT:

    - ``complement`` -- boolean (default: ``True``); whether to use Godsil's duality
      theorem to compute the matching polynomial from that of the graphs
      complement (see ALGORITHM)

    - ``name`` -- (optional) string for the variable name in the polynomial

    .. NOTE::

        The ``complement`` option uses matching polynomials of complete graphs,
        which are cached. So if you are crazy enough to try computing the
        matching polynomial on a graph with millions of vertices, you might not
        want to use this option, since it will end up caching millions of
        polynomials of degree in the millions.

    ALGORITHM:

    The algorithm used is a recursive one, based on the following observation
    [God1993]_:

    - If `e` is an edge of `G`, `G'` is the result of deleting the edge `e`, and
      `G''` is the result of deleting each vertex in `e`, then the matching
      polynomial of `G` is equal to that of `G'` minus that of `G''`.

      (the algorithm actually computes the *signless* matching polynomial, for
      which the recursion is the same when one replaces the subtraction by an
      addition. It is then converted into the matching polynomial and returned)

    Depending on the value of ``complement``, Godsil's duality theorem
    [God1993]_ can also be used to compute `\mu(x)` :

    .. MATH::

        \mu(\overline{G}, x) = \sum_{k \geq 0} p(G,k) \mu( K_{n-2k}, x)


    Where `\overline{G}` is the complement of `G`, and `K_n` the complete graph
    on `n` vertices.

    EXAMPLES::

        sage: g = graphs.PetersenGraph()
        sage: g.matching_polynomial()
        x^10 - 15*x^8 + 75*x^6 - 145*x^4 + 90*x^2 - 6
        sage: g.matching_polynomial(complement=False)
        x^10 - 15*x^8 + 75*x^6 - 145*x^4 + 90*x^2 - 6
        sage: g.matching_polynomial(name='tom')
        tom^10 - 15*tom^8 + 75*tom^6 - 145*tom^4 + 90*tom^2 - 6
        sage: g = Graph()
        sage: L = [graphs.RandomGNP(8, .3) for i in range(1, 6)]
        sage: prod([h.matching_polynomial() for h in L]) == sum(L, g).matching_polynomial()  # long time (up to 10s on sage.math, 2011)
        True

    ::

        sage: for i in range(1, 12):  # long time (10s on sage.math, 2011)
        ....:     for t in graphs.trees(i):
        ....:         if t.matching_polynomial() != t.characteristic_polynomial():
        ....:             raise RuntimeError('bug for a tree A of size {0}'.format(i))
        ....:         c = t.complement()
        ....:         if c.matching_polynomial(complement=False) != c.matching_polynomial():
        ....:             raise RuntimeError('bug for a tree B of size {0}'.format(i))

    ::

        sage: from sage.graphs.matchpoly import matching_polynomial
        sage: matching_polynomial(graphs.CompleteGraph(0))
        1
        sage: matching_polynomial(graphs.CompleteGraph(1))
        x
        sage: matching_polynomial(graphs.CompleteGraph(2))
        x^2 - 1
        sage: matching_polynomial(graphs.CompleteGraph(3))
        x^3 - 3*x
        sage: matching_polynomial(graphs.CompleteGraph(4))
        x^4 - 6*x^2 + 3
        sage: matching_polynomial(graphs.CompleteGraph(5))
        x^5 - 10*x^3 + 15*x
        sage: matching_polynomial(graphs.CompleteGraph(6))
        x^6 - 15*x^4 + 45*x^2 - 15
        sage: matching_polynomial(graphs.CompleteGraph(7))
        x^7 - 21*x^5 + 105*x^3 - 105*x
        sage: matching_polynomial(graphs.CompleteGraph(8))
        x^8 - 28*x^6 + 210*x^4 - 420*x^2 + 105
        sage: matching_polynomial(graphs.CompleteGraph(9))
        x^9 - 36*x^7 + 378*x^5 - 1260*x^3 + 945*x
        sage: matching_polynomial(graphs.CompleteGraph(10))
        x^10 - 45*x^8 + 630*x^6 - 3150*x^4 + 4725*x^2 - 945
        sage: matching_polynomial(graphs.CompleteGraph(11))
        x^11 - 55*x^9 + 990*x^7 - 6930*x^5 + 17325*x^3 - 10395*x
        sage: matching_polynomial(graphs.CompleteGraph(12))
        x^12 - 66*x^10 + 1485*x^8 - 13860*x^6 + 51975*x^4 - 62370*x^2 + 10395
        sage: matching_polynomial(graphs.CompleteGraph(13))
        x^13 - 78*x^11 + 2145*x^9 - 25740*x^7 + 135135*x^5 - 270270*x^3 + 135135*x

    ::

        sage: G = Graph({0:[1,2], 1:[2]})
        sage: matching_polynomial(G)
        x^3 - 3*x
        sage: G = Graph({0:[1,2]})
        sage: matching_polynomial(G)
        x^3 - 2*x
        sage: G = Graph({0:[1], 2:[]})
        sage: matching_polynomial(G)
        x^3 - x
        sage: G = Graph({0:[], 1:[], 2:[]})
        sage: matching_polynomial(G)
        x^3

    ::

        sage: matching_polynomial(graphs.CompleteGraph(0), complement=False)
        1
        sage: matching_polynomial(graphs.CompleteGraph(1), complement=False)
        x
        sage: matching_polynomial(graphs.CompleteGraph(2), complement=False)
        x^2 - 1
        sage: matching_polynomial(graphs.CompleteGraph(3), complement=False)
        x^3 - 3*x
        sage: matching_polynomial(graphs.CompleteGraph(4), complement=False)
        x^4 - 6*x^2 + 3
        sage: matching_polynomial(graphs.CompleteGraph(5), complement=False)
        x^5 - 10*x^3 + 15*x
        sage: matching_polynomial(graphs.CompleteGraph(6), complement=False)
        x^6 - 15*x^4 + 45*x^2 - 15
        sage: matching_polynomial(graphs.CompleteGraph(7), complement=False)
        x^7 - 21*x^5 + 105*x^3 - 105*x
        sage: matching_polynomial(graphs.CompleteGraph(8), complement=False)
        x^8 - 28*x^6 + 210*x^4 - 420*x^2 + 105
        sage: matching_polynomial(graphs.CompleteGraph(9), complement=False)
        x^9 - 36*x^7 + 378*x^5 - 1260*x^3 + 945*x
        sage: matching_polynomial(graphs.CompleteGraph(10), complement=False)
        x^10 - 45*x^8 + 630*x^6 - 3150*x^4 + 4725*x^2 - 945
        sage: matching_polynomial(graphs.CompleteGraph(11), complement=False)
        x^11 - 55*x^9 + 990*x^7 - 6930*x^5 + 17325*x^3 - 10395*x
        sage: matching_polynomial(graphs.CompleteGraph(12), complement=False)
        x^12 - 66*x^10 + 1485*x^8 - 13860*x^6 + 51975*x^4 - 62370*x^2 + 10395
        sage: matching_polynomial(graphs.CompleteGraph(13), complement=False)
        x^13 - 78*x^11 + 2145*x^9 - 25740*x^7 + 135135*x^5 - 270270*x^3 + 135135*x

    TESTS:

    Non-integer labels should work, (:issue:`15545`)::

        sage: G = Graph(10)
        sage: G.add_vertex((0,1))
        sage: G.add_vertex('X')
        sage: G.matching_polynomial()
        x^12

    Check that :issue:`39930` is fixed::

        sage: G = Graph([(0, 1), (1, 2), (2, 2)], immutable=True, loops=True)
        sage: G.matching_polynomial()
        x^3 - 3*x
    """
    if G.has_multiple_edges():
        raise NotImplementedError

    cdef int i, j, d
    cdef fmpz_poly_t pol
    cdef int nverts = G.n_vertices()

    # Using Godsil's duality theorem when the graph is dense

    if complement and G.density() > 0.5:  # this cutoff could probably be tuned
        f_comp = matching_polynomial(G.complement()).list()
        f = x.parent().zero()
        for i in range(nverts // 2 + 1):
            j = nverts - 2 * i
            f += complete_poly(j) * f_comp[j] * (-1)**i
        return f

    cdef int nedges = G.n_edges()

    # Relabelling the vertices of the graph as [0...n-1] so that they are sorted
    # in increasing order of degree

    cdef list L = [(d, v) for v, d in G.degree_iterator(labels=True)]
    L.sort(key=lambda pair: pair[0])
    Gint = G.relabel(perm={v: i for i, (_, v) in enumerate(L)},
                     inplace=False, immutable=False)
    Gint.allow_loops(False)

    # Initialization of pol, edges* variables.

    # The edges_mem table is of size (2 * nedges * nedges), and is to be read as
    # nedges blocks of size (2 * nedges). These blocks of size (2 * nedges) are
    # themselves to be read as two blocks of length nedges, addressed as edges1
    # and edges2.
    #
    # Only the first block of size (2 * nedges) is here filled. The function
    # delete_and_add will need the rest of the memory.

    fmpz_poly_init(pol)  # sets to zero
    cdef int** edges = <int**>check_allocarray(2 * nedges, sizeof(int*))
    cdef int* edges_mem = <int*>check_allocarray(2 * nedges * nedges, sizeof(int))

    for i in range(2 * nedges):
        edges[i] = edges_mem + i * nedges

    cdef int* edges1 = edges_mem           # edges[0]
    cdef int* edges2 = edges_mem + nedges  # edges[1]

    cdef int cur = 0
    for i, j in Gint.edges(labels=False, sort=True, sort_vertices=True):
        edges1[cur] = i
        edges2[cur] = j
        cur += 1

    # Computing the signless matching polynomial

    sig_on()
    delete_and_add(edges, nverts, nedges, nverts, 0, pol)
    sig_off()

    # Building the actual matching polynomial

    cdef list coeffs_ZZ = []
    cdef Integer c_ZZ
    for i in range(nverts + 1):
        c_ZZ = Integer(0)
        fmpz_poly_get_coeff_mpz(c_ZZ.value, pol, i)
        coeffs_ZZ.append(c_ZZ * (-1)**((nverts - i) // 2))

    f = x.parent()(coeffs_ZZ)
    sig_free(edges_mem)
    sig_free(edges)
    fmpz_poly_clear(pol)
    if name is not None:
        return f.change_variable_name(name)
    return f


# The following is a cache of complete graph matching polynomials.

cdef list complete_matching_polys = [x.parent().one(), x]


def complete_poly(n):
    """
    Compute the matching polynomial of the complete graph on `n` vertices.

    INPUT:

    - ``n`` -- order of the complete graph

    .. TODO::

        This code could probably be made more efficient by using FLINT
        polynomials and being written in Cython, using an array of
        fmpz_poly_t pointers or something...  Right now just about the
        whole complement optimization is written in Python, and could
        be easily sped up.

    EXAMPLES::

        sage: from sage.graphs.matchpoly import complete_poly
        sage: f = complete_poly(10)
        sage: f
        x^10 - 45*x^8 + 630*x^6 - 3150*x^4 + 4725*x^2 - 945
        sage: f = complete_poly(20)
        sage: f[8]
        1309458150
        sage: f = complete_poly(1000)
        sage: len(str(f))
        406824

    TESTS:

    Checking the numerical results up to 20::

        sage: from sage.functions.orthogonal_polys import hermite                       # needs sage.symbolic
        sage: p = lambda n: 2^(-n/2)*hermite(n, x/sqrt(2))                              # needs sage.symbolic
        sage: all(p(i) == complete_poly(i) for i in range(2, 20))                       # needs sage.symbolic
        True
    """
    # global complete_matching_polys # if we do eventually make it a C array...
    cdef int lcmp = len(complete_matching_polys)

    if n < lcmp:
        return complete_matching_polys[n]
    lcmp -= 2
    a = complete_matching_polys[lcmp]
    b = complete_matching_polys[lcmp + 1]
    while lcmp + 2 <= n:
        a, b, lcmp = b, x * b - (lcmp + 1) * a, lcmp + 1
        complete_matching_polys.append(b)
    return b


cdef void delete_and_add(int **edges, int nverts, int nedges, int totverts, int depth, fmpz_poly_t pol) noexcept:
    """
    Add matching polynomial to pol via recursion.

    NOTE : at the end of this function, pol represents the *SIGNLESS*
    matching polynomial.
    """
    cdef fmpz* coeff

    if nverts == 3:
        coeff = fmpz_poly_get_coeff_ptr(pol, 3)
        if coeff is NULL:
            fmpz_poly_set_coeff_ui(pol, 3, 1)
        else:
            fmpz_add_ui(coeff, coeff, 1)
        coeff = fmpz_poly_get_coeff_ptr(pol, 1)
        if coeff is NULL:
            fmpz_poly_set_coeff_ui(pol, 1, nedges)
        else:
            fmpz_add_ui(coeff, coeff, nedges)
        return

    if not nedges:
        coeff = fmpz_poly_get_coeff_ptr(pol, nverts)
        if coeff is NULL:
            fmpz_poly_set_coeff_ui(pol, nverts, 1)
        else:
            fmpz_add_ui(coeff, coeff, 1)
        return

    cdef int* edges1 = edges[2 * depth]
    cdef int* edges2 = edges[2 * depth + 1]
    cdef int* new_edges1 = edges[2 * depth + 2]
    cdef int* new_edges2 = edges[2 * depth + 3]

    nedges -= 1

    # The last edge is (edge1, edge2)
    cdef int edge1 = edges1[nedges]
    cdef int edge2 = edges2[nedges]
    cdef int new_nedges = 0
    cdef int i

    # The new edges are all the edges that are not incident with (edge1, edge2)

    for i in range(nedges):
        if edge1 == edges1[i]:
            # Since the rest of the edges are incident to edge1 (the edges are
            # sorted by increasing order of their first component), we can stop
            break

        if edge1 != edges2[i] and edge2 != edges2[i]:
            # since edge2 > edge1 > edges1[i], we only need to check edges2[i]
            new_edges1[new_nedges] = edges1[i]
            new_edges2[new_nedges] = edges2[i]
            new_nedges += 1

    delete_and_add(edges, nverts - 2, new_nedges, totverts, depth + 1, pol)
    delete_and_add(edges, nverts, nedges, totverts, depth, pol)



"""
Interacts for Calculus

AUTHORS:

- Harald Schilly (2011-01-16): initial version (#9623) partially based on work by Lauri Ruotsalainen
"""

# ****************************************************************************
#       Copyright (C) 2011 Harald Schilly <harald.schilly@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from .library import taylor_polynomial
from .library import definite_integral
from .library import function_derivative, difference_quotient, quadratic_equation
from .library import trigonometric_properties_triangle
from .library import secant_method, newton_method, trapezoid_integration, simpson_integration, bisection_method
from .library import riemann_sum
from .library import function_tool



# distutils: libraries = NTL_LIBRARIES gmp m
# distutils: extra_compile_args = NTL_CFLAGS
# distutils: include_dirs = NTL_INCDIR
# distutils: library_dirs = NTL_LIBDIR
# distutils: extra_link_args = NTL_LIBEXTRA
# distutils: language = c++

# ****************************************************************************
#       Copyright (C) 2005 William Stein <wstein@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from cysignals.signals cimport sig_on, sig_off

from sage.cpython.string cimport char_to_str
from sage.ext.cplusplus cimport ccreadstr

include "decl.pxi"
include 'misc.pxi'

from cpython.object cimport Py_EQ, Py_NE
from sage.libs.ntl.ntl_ZZ cimport ntl_ZZ
from sage.libs.ntl.ntl_ZZ import unpickle_class_value

from sage.rings.integer_ring import IntegerRing
from sage.arith.power cimport generic_power_pos

ZZ = IntegerRing()

cdef inline ntl_ZZ make_ZZ(ZZ_c* x):
    """ These make_XXXX functions are deprecated and should be phased out."""
    cdef ntl_ZZ y
    y = ntl_ZZ()
    y.x = x[0]
    del x
    return y

# You must do sig_on() before calling this function
cdef inline ntl_ZZ make_ZZ_sig_off(ZZ_c* x):
    cdef ntl_ZZ y = make_ZZ(x)
    sig_off()
    return y

cdef inline ntl_ZZX make_ZZX(ZZX_c* x):
    """ These make_XXXX functions are deprecated and should be phased out."""
    cdef ntl_ZZX y
    y = ntl_ZZX()
    y.x = x[0]
    del x
    return y

# You must do sig_on() before calling this function
cdef inline ntl_ZZX make_ZZX_sig_off(ZZX_c* x):
    cdef ntl_ZZX y = make_ZZX(x)
    sig_off()
    return y

from sage.structure.proof.proof import get_flag
cdef proof_flag(t):
    return get_flag(t, "polynomial")

##############################################################################
#
# ZZX: polynomials over the integers
#
##############################################################################


cdef class ntl_ZZX():
    r"""
    The class \class{ZZX} implements polynomials in `\Z[X]`, i.e.,
    univariate polynomials with integer coefficients.

    Polynomial multiplication is very fast, and is implemented using
    one of 4 different algorithms:
    \begin{enumerate}
    \item\hspace{1em} Classical
    \item\hspace{1em} Karatsuba
    \item\hspace{1em} Schoenhage and Strassen --- performs an FFT by working
          modulo a "Fermat number" of appropriate size...
          good for polynomials with huge coefficients
          and moderate degree
    \item\hspace{1em} CRT/FFT --- performs an FFT by working modulo several
         small primes.  This is good for polynomials with moderate
         coefficients and huge degree.
    \end{enumerate}

    The choice of algorithm is somewhat heuristic, and may not always be
    perfect.

    Many thanks to Juergen Gerhard {\tt
    <jngerhar@plato.uni-paderborn.de>} for pointing out the deficiency
    in the NTL-1.0 ZZX arithmetic, and for contributing the
    Schoenhage/Strassen code.

    Extensive use is made of modular algorithms to enhance performance
    (e.g., the GCD algorithm and many others).
    """

    # See ntl_ZZX.pxd for definition of data members
    def __init__(self, v=None):
        """
        EXAMPLES::

            sage: f = ntl.ZZX([1,2,5,-9])
            sage: f
            [1 2 5 -9]
            sage: g = ntl.ZZX([0,0,0]); g
            []
            sage: g[10]=5
            sage: g
            [0 0 0 0 0 0 0 0 0 0 5]
            sage: g[10]
            5
        """
        cdef ntl_ZZ cc
        cdef Py_ssize_t i

        if v is None:
            return
        elif isinstance(v, (list, tuple)):
            for i from 0 <= i < len(v):
                x = v[i]
                if not isinstance(x, ntl_ZZ):
                    cc = ntl_ZZ(x)
                else:
                    cc = x
                ZZX_SetCoeff(self.x, i, cc.x)
        else:
            ccreadstr(self.x, str(v))

    def __reduce__(self):
        """
        EXAMPLES::

            sage: from sage.libs.ntl.ntl_ZZX import ntl_ZZX
            sage: f = ntl_ZZX([1,2,0,4])
            sage: loads(dumps(f)) == f
            True
        """
        return unpickle_class_value, (ntl_ZZX, self.list())

    def __repr__(self):
        """
        Return the string representation of ``self``.

        EXAMPLES::

            sage: ntl.ZZX([1,3,0,5]).__repr__()
            '[1 3 0 5]'
        """
        cdef char * val
        val = ZZX_repr(&self.x)
        result = char_to_str(val)
        cpp_delete_array(val)
        return result

    def __copy__(self):
        """
        Return a copy of ``self``.

        EXAMPLES::

            sage: x = ntl.ZZX([1,32])
            sage: y = copy(x)
            sage: y == x
            True
            sage: y is x
            False
        """
        return make_ZZX(ZZX_copy(&self.x))

    def __setitem__(self, long i, a):
        """
        EXAMPLES::

            sage: n=ntl.ZZX([1,2,3])
            sage: n
            [1 2 3]
            sage: n[1] = 4
            sage: n
            [1 4 3]
        """
        if i < 0:
            raise IndexError("index (i=%s) must be >= 0" % i)
        cdef ntl_ZZ cc
        if isinstance(a, ntl_ZZ):
            cc = a
        else:
            cc = ntl_ZZ(a)
        ZZX_SetCoeff(self.x, i, cc.x)

    cdef void setitem_from_int(ntl_ZZX self, long i, int value) noexcept:
        r"""
        Set `i`-th coefficient to value.

        AUTHOR: David Harvey (2006-08-05)
        """
        ZZX_setitem_from_int(&self.x, i, value)

    def setitem_from_int_doctest(self, i, value):
        r"""
        This method exists solely for automated testing of setitem_from_int().

        sage: x = ntl.ZZX([2, 3, 4])
        sage: x.setitem_from_int_doctest(5, 42)
        sage: x
         [2 3 4 0 0 42]
        """
        self.setitem_from_int(int(i), int(value))

    def __getitem__(self, long i):
        r"""
        Retrieves coefficient number i as an NTL ZZ.

        sage: x = ntl.ZZX([129381729371289371237128318293718237, 2, -3, 0, 4])
        sage: x[0]
         129381729371289371237128318293718237
        sage: type(x[0])
         <class 'sage.libs.ntl.ntl_ZZ.ntl_ZZ'>
        sage: x[1]
         2
        sage: x[2]
         -3
        sage: x[3]
         0
        sage: x[4]
         4
        sage: x[5]
         0
        """
        cdef ntl_ZZ r = ntl_ZZ()
        sig_on()
        r.x = ZZX_coeff(self.x, <long>i)
        sig_off()
        return r

    cdef int getitem_as_int(ntl_ZZX self, long i) noexcept:
        r"""
        Return `i`-th coefficient as C int.
        Return value is only valid if the result fits into an int.

        AUTHOR: David Harvey (2006-08-05)
        """
        return ZZX_getitem_as_int(&self.x, i)

    def getitem_as_int_doctest(self, i):
        r"""
        This method exists solely for automated testing of getitem_as_int().

        sage: x = ntl.ZZX([2, 3, 5, -7, 11])
        sage: i = x.getitem_as_int_doctest(3)
        sage: i
         -7
        sage: type(i)
         <... 'int'>
        sage: x.getitem_as_int_doctest(15)
         0
        """
        return self.getitem_as_int(i)

    def list(self):
        r"""
        Retrieves coefficients as a list of ntl.ZZ Integers.

        EXAMPLES::

            sage: x = ntl.ZZX([129381729371289371237128318293718237, 2, -3, 0, 4])
            sage: L = x.list(); L
            [129381729371289371237128318293718237, 2, -3, 0, 4]
            sage: type(L[0])
            <class 'sage.libs.ntl.ntl_ZZ.ntl_ZZ'>
            sage: x = ntl.ZZX()
            sage: L = x.list(); L
            []
        """
        cdef int i
        return [self[i] for i from 0 <= i <= ZZX_deg(self.x)]

    def __add__(ntl_ZZX self, ntl_ZZX other):
        """
        EXAMPLES::

            sage: ntl.ZZX(list(range(5))) + ntl.ZZX(list(range(6)))
            [0 2 4 6 8 5]
        """
        cdef ntl_ZZX r = ntl_ZZX.__new__(ntl_ZZX)
        if not isinstance(self, ntl_ZZX):
            self = ntl_ZZX(self)
        if not isinstance(other, ntl_ZZX):
            other = ntl_ZZX(other)
        ZZX_add(r.x, (<ntl_ZZX>self).x, (<ntl_ZZX>other).x)
        return r

    def __sub__(ntl_ZZX self, ntl_ZZX other):
        """
        EXAMPLES::

            sage: ntl.ZZX(list(range(5))) - ntl.ZZX(list(range(6)))
            [0 0 0 0 0 -5]
        """
        cdef ntl_ZZX r = ntl_ZZX.__new__(ntl_ZZX)
        if not isinstance(self, ntl_ZZX):
            self = ntl_ZZX(self)
        if not isinstance(other, ntl_ZZX):
            other = ntl_ZZX(other)
        ZZX_sub(r.x, (<ntl_ZZX>self).x, (<ntl_ZZX>other).x)
        return r

    def __mul__(ntl_ZZX self, ntl_ZZX other):
        """
        EXAMPLES::

            sage: ntl.ZZX(list(range(5))) * ntl.ZZX(list(range(6)))
            [0 0 1 4 10 20 30 34 31 20]
        """
        cdef ntl_ZZX r = ntl_ZZX.__new__(ntl_ZZX)
        if not isinstance(self, ntl_ZZX):
            self = ntl_ZZX(self)
        if not isinstance(other, ntl_ZZX):
            other = ntl_ZZX(other)
        sig_on()
        ZZX_mul(r.x, (<ntl_ZZX>self).x, (<ntl_ZZX>other).x)
        sig_off()
        return r

    def __truediv__(ntl_ZZX self, ntl_ZZX other):
        """
        Compute quotient ``self / other``, if the quotient is a polynomial.
        Otherwise an Exception is raised.

        EXAMPLES::

            sage: f = ntl.ZZX([1,2,3]) * ntl.ZZX([4,5])**2
            sage: g = ntl.ZZX([4,5])
            sage: f/g
            [4 13 22 15]
            sage: ntl.ZZX([1,2,3]) * ntl.ZZX([4,5])
            [4 13 22 15]

            sage: f = ntl.ZZX(list(range(10))); g = ntl.ZZX([-1,0,1])
            sage: f/g
            Traceback (most recent call last):
            ...
            ArithmeticError: self (=[0 1 2 3 4 5 6 7 8 9]) is not divisible by other (=[-1 0 1])
        """
        sig_on()
        cdef int divisible
        cdef ZZX_c* q
        q = ZZX_div(&self.x, &other.x, &divisible)
        if not divisible:
            del q
            sig_off()
            raise ArithmeticError("self (=%s) is not divisible by other (=%s)" % (self, other))
        result = make_ZZX_sig_off(q)
        return result

    def __mod__(ntl_ZZX self, ntl_ZZX other):
        """
        Given polynomials a, b in ZZ[X], there exist polynomials q, r
        in QQ[X] such that a = b*q + r, deg(r) < deg(b).  This
        function returns q if q lies in ZZ[X], and otherwise raises an
        Exception.

        EXAMPLES::

            sage: f = ntl.ZZX([2,4,6]); g = ntl.ZZX([2])
            sage: f % g   # 0
            []

            sage: f = ntl.ZZX(list(range(10))); g = ntl.ZZX([-1,0,1])
            sage: f % g
            [20 25]
        """
        cdef ntl_ZZX r = ntl_ZZX.__new__(ntl_ZZX)
        if not isinstance(self, ntl_ZZX):
            self = ntl_ZZX(self)
        if not isinstance(other, ntl_ZZX):
            other = ntl_ZZX(other)
        sig_on()
        ZZX_rem(r.x, (<ntl_ZZX>self).x, (<ntl_ZZX>other).x)
        sig_off()
        return r

    def quo_rem(self, ntl_ZZX other):
        """
        Return the unique integral q and r such that ``self = q*other +
        r``, if they exist.  Otherwise raises an Exception.

        EXAMPLES::

           sage: f = ntl.ZZX(list(range(10))); g = ntl.ZZX([-1,0,1])
           sage: q, r = f.quo_rem(g)
           sage: q, r
           ([20 24 18 21 14 16 8 9], [20 25])
           sage: q*g + r == f
           True
        """
        cdef ZZX_c *r
        cdef ZZX_c *q
        sig_on()
        try:
            ZZX_quo_rem(&self.x, &other.x, &r, &q)
            return (make_ZZX(q), make_ZZX(r))
        finally:
            sig_off()

    def square(self):
        """
        Return f*f.

        EXAMPLES::

            sage: f = ntl.ZZX([-1,0,1])
            sage: f*f
            [1 0 -2 0 1]
        """
        sig_on()
        return make_ZZX_sig_off(ZZX_square(&self.x))

    def __pow__(ntl_ZZX self, long n, ignored):
        """
        Return the `n`-th nonnegative power of ``self``.

        EXAMPLES::

            sage: g = ntl.ZZX([-1,0,1])
            sage: g ^ 10
            [1 0 -10 0 45 0 -120 0 210 0 -252 0 210 0 -120 0 45 0 -10 0 1]
            sage: g ^ 0
            [1]
            sage: g ^ 1
            [-1 0 1]
            sage: g ^ (-1)
            Traceback (most recent call last):
            ...
            ArithmeticError
        """
        if n == 0:
            from copy import copy
            return copy(one_ZZX)
        if n < 0:
            raise ArithmeticError
        return generic_power_pos(self, <unsigned long>n)

    def __richcmp__(ntl_ZZX self, other, int op):
        """
        Compare ``self`` to ``other``.

        EXAMPLES::

            sage: f = ntl.ZZX([1,2,3])
            sage: g = ntl.ZZX([1,2,3,0])
            sage: f == g
            True
            sage: g = ntl.ZZX([0,1,2,3])
            sage: f == g
            False
            sage: f == ntl.ZZ(0)
            False
        """
        if op != Py_EQ and op != Py_NE:
            raise TypeError("polynomials are not ordered")

        cdef ntl_ZZX b
        try:
            b = <ntl_ZZX?>other
        except TypeError:
            b = ntl_ZZX(other)

        return (op == Py_EQ) == (self.x == b.x)

    def is_zero(self):
        """
        Return ``True`` exactly if this polynomial is 0.

        EXAMPLES::

            sage: f = ntl.ZZX([0,0,0,0])
            sage: f.is_zero()
            True
            sage: f = ntl.ZZX([0,0,1])
            sage: f
            [0 0 1]
            sage: f.is_zero()
            False
        """
        return bool(ZZX_IsZero(self.x))

    def is_one(self):
        """
        Return ``True`` exactly if this polynomial is 1.

        EXAMPLES::

            sage: f = ntl.ZZX([1,1])
            sage: f.is_one()
            False
            sage: f = ntl.ZZX([1])
            sage: f.is_one()
            True
        """
        return bool(ZZX_IsOne(self.x))

    def is_monic(self):
        """
        Return ``True`` exactly if this polynomial is monic.

        EXAMPLES::

            sage: f = ntl.ZZX([2,0,0,1])
            sage: f.is_monic()
            True
            sage: g = f.reverse()
            sage: g.is_monic()
            False
            sage: g
            [1 0 0 2]
        """
        if ZZX_IsZero(self.x):
            return False
        cdef ZZ_c lc
        lc = ZZX_LeadCoeff(self.x)
        return <bint>ZZ_IsOne(lc)

        # return bool(ZZX_is_monic(&self.x))

    def __neg__(self):
        """
        Return the negative of ``self``.

        EXAMPLES::

            sage: f = ntl.ZZX([2,0,0,1])
            sage: -f
            [-2 0 0 -1]
        """
        return make_ZZX(ZZX_neg(&self.x))

    def left_shift(self, long n):
        """
        Return the polynomial obtained by shifting all coefficients of
        this polynomial to the left n positions.

        EXAMPLES::

            sage: f = ntl.ZZX([2,0,0,1])
            sage: f
            [2 0 0 1]
            sage: f.left_shift(2)
            [0 0 2 0 0 1]
            sage: f.left_shift(5)
            [0 0 0 0 0 2 0 0 1]

        A negative left shift is a right shift.
            sage: f.left_shift(-2)
            [0 1]
        """
        return make_ZZX(ZZX_left_shift(&self.x, n))

    def right_shift(self, long n):
        """
        Return the polynomial obtained by shifting all coefficients of
        this polynomial to the right n positions.

        EXAMPLES::

            sage: f = ntl.ZZX([2,0,0,1])
            sage: f
            [2 0 0 1]
            sage: f.right_shift(2)
            [0 1]
            sage: f.right_shift(5)
            []
            sage: f.right_shift(-2)
            [0 0 2 0 0 1]
        """
        return make_ZZX(ZZX_right_shift(&self.x, n))

    def content(self):
        """
        Return the content of f, which has sign the same as the
        leading coefficient of f.  Also, our convention is that the
        content of 0 is 0.

        EXAMPLES::

            sage: f = ntl.ZZX([2,0,0,2])
            sage: f.content()
            2
            sage: f = ntl.ZZX([2,0,0,-2])
            sage: f.content()
            -2
            sage: f = ntl.ZZX([6,12,3,9])
            sage: f.content()
            3
            sage: f = ntl.ZZX([])
            sage: f.content()
            0
        """
        cdef ntl_ZZ r = ntl_ZZ.__new__(ntl_ZZ)
        ZZX_content(r.x, self.x)
        return r

    def primitive_part(self):
        """
        Return the primitive part of f.  Our convention is that the leading
        coefficient of the primitive part is nonnegative, and the primitive
        part of 0 is 0.

        EXAMPLES::

            sage: f = ntl.ZZX([6,12,3,9])
            sage: f.primitive_part()
            [2 4 1 3]
            sage: f
            [6 12 3 9]
            sage: f = ntl.ZZX([6,12,3,-9])
            sage: f
            [6 12 3 -9]
            sage: f.primitive_part()
            [-2 -4 -1 3]
            sage: f = ntl.ZZX()
            sage: f.primitive_part()
            []
        """
        return make_ZZX(ZZX_primitive_part(&self.x))

    def pseudo_quo_rem(self, ntl_ZZX other):
        r"""
        Perform pseudo-division: computes q and r with deg(r) <
        deg(b), and \code{LeadCoeff(b)\^(deg(a)-deg(b)+1) a = b q + r}.
        Only the classical algorithm is used.

        EXAMPLES::

            sage: f = ntl.ZZX([0,1])
            sage: g = ntl.ZZX([1,2,3])
            sage: g.pseudo_quo_rem(f)
            ([2 3], [1])
            sage: f = ntl.ZZX([1,1])
            sage: g.pseudo_quo_rem(f)
            ([-1 3], [2])
        """
        cdef ZZX_c *r
        cdef ZZX_c *q
        sig_on()
        try:
            ZZX_pseudo_quo_rem(&self.x, &other.x, &r, &q)
            return (make_ZZX(q), make_ZZX(r))
        finally:
            sig_off()

    def gcd(self, ntl_ZZX other):
        """
        Return the gcd d = gcd(a, b), where by convention the leading coefficient
        of d is >= 0.  We use a multi-modular algorithm.

        EXAMPLES::

            sage: f = ntl.ZZX([1,2,3]) * ntl.ZZX([4,5])**2
            sage: g = ntl.ZZX([1,1,1])**3 * ntl.ZZX([1,2,3])
            sage: f.gcd(g)
            [1 2 3]
            sage: g.gcd(f)
            [1 2 3]
        """
        sig_on()
        return make_ZZX_sig_off(ZZX_gcd(&self.x, &other.x))

    def lcm(self, ntl_ZZX other):
        """
        Return the least common multiple of ``self`` and ``other``.

        EXAMPLES::

            sage: x1 = ntl.ZZX([-1,0,0,1])
            sage: x2 = ntl.ZZX([-1,0,0,0,0,0,1])
            sage: x1.lcm(x2)
            [-1 0 0 0 0 0 1]
        """
        g = self.gcd(other)
        return (self*other).quo_rem(g)[0]

    def xgcd(self, ntl_ZZX other, proof=None):
        r"""
        If ``self`` and ``other`` are coprime over the rationals,
        return ``r, s, t`` such that ``r = s*self + t*other``.
        Otherwise return 0.

        This is \emph{not} the same as the \sage function on
        polynomials over the integers, since here the return value r
        is always an integer.

        Here r is the resultant of a and b; if r != 0, then this
        function computes s and t such that: a*s + b*t = r; otherwise
        s and t are both 0.  If proof = False (*not* the default),
        then resultant computation may use a randomized strategy that
        errors with probability no more than `2^{-80}`.  The default is
        default is proof=None, see proof.polynomial or sage.structure.proof,
        but the global default is True), then this function may use a
        randomized strategy that errors with probability no more than
        `2^{-80}`.

        EXAMPLES::

            sage: f = ntl.ZZX([1,2,3]) * ntl.ZZX([4,5])**2
            sage: g = ntl.ZZX([1,1,1])**3 * ntl.ZZX([1,2,3])
            sage: f.xgcd(g)   # nothing since they are not coprime
            (0, [], [])

        In this example the input quadratic polynomials have a common root modulo 13::

            sage: f = ntl.ZZX([5,0,1])
            sage: g = ntl.ZZX([18,0,1])
            sage: f.xgcd(g)
            (169, [-13], [13])
        """
        proof = proof_flag(proof)

        cdef ZZX_c *s
        cdef ZZX_c *t
        cdef ZZ_c *r
        sig_on()
        try:
            ZZX_xgcd(&self.x, &other.x, &r, &s, &t, proof)
            return (make_ZZ(r), make_ZZX(s), make_ZZX(t))
        finally:
            sig_off()

    def degree(self):
        """
        Return the degree of this polynomial.  The degree of the 0
        polynomial is -1.

        EXAMPLES::

            sage: f = ntl.ZZX([5,0,1])
            sage: f.degree()
            2
            sage: f = ntl.ZZX(list(range(100)))
            sage: f.degree()
            99
            sage: f = ntl.ZZX()
            sage: f.degree()
            -1
            sage: f = ntl.ZZX([1])
            sage: f.degree()
            0
        """
        return ZZX_deg(self.x)

    def leading_coefficient(self):
        """
        Return the leading coefficient of this polynomial.

        EXAMPLES::

            sage: f = ntl.ZZX([3,6,9])
            sage: f.leading_coefficient()
            9
            sage: f = ntl.ZZX()
            sage: f.leading_coefficient()
            0
        """
        cdef ntl_ZZ r = ntl_ZZ.__new__(ntl_ZZ)
        r.x = ZZX_LeadCoeff(self.x)
        return r

    def constant_term(self):
        """
        Return the constant coefficient of this polynomial.

        EXAMPLES::

            sage: f = ntl.ZZX([3,6,9])
            sage: f.constant_term()
            3
            sage: f = ntl.ZZX()
            sage: f.constant_term()
            0
        """
        cdef ntl_ZZ r = ntl_ZZ.__new__(ntl_ZZ)
        r.x = ZZX_ConstTerm(self.x)
        return r

    def set_x(self):
        """
        Set this polynomial to the monomial "x".

        EXAMPLES::

            sage: f = ntl.ZZX()
            sage: f.set_x()
            sage: f
            [0 1]
            sage: g = ntl.ZZX([0,1])
            sage: f == g
            True

        Though f and g are equal, they are not the same objects in memory::

            sage: f is g
            False
        """
        ZZX_set_x(&self.x)

    def is_x(self):
        """
        ``True`` if this is the polynomial "x".

        EXAMPLES::

            sage: f = ntl.ZZX()
            sage: f.set_x()
            sage: f.is_x()
            True
            sage: f = ntl.ZZX([0,1])
            sage: f.is_x()
            True
            sage: f = ntl.ZZX([1])
            sage: f.is_x()
            False
        """
        return bool(ZZX_is_x(&self.x))

    def derivative(self):
        """
        Return the derivative of this polynomial.

        EXAMPLES::

            sage: f = ntl.ZZX([1,7,0,13])
            sage: f.derivative()
            [7 0 39]
        """
        return make_ZZX(ZZX_derivative(&self.x))

    def reverse(self, hi=None):
        """
        Return the polynomial obtained by reversing the coefficients
        of this polynomial.  If hi is set then this function behaves
        as if this polynomial has degree hi.

        EXAMPLES::

            sage: f = ntl.ZZX([1,2,3,4,5])
            sage: f.reverse()
            [5 4 3 2 1]
            sage: f.reverse(hi=10)
            [0 0 0 0 0 0 5 4 3 2 1]
            sage: f.reverse(hi=2)
            [3 2 1]
            sage: f.reverse(hi=-2)
            []
        """
        if not (hi is None):
            return make_ZZX(ZZX_reverse_hi(&self.x, int(hi)))
        else:
            return make_ZZX(ZZX_reverse(&self.x))

    def truncate(self, long m):
        """
        Return the truncation of this polynomial obtained by
        removing all terms of degree >= m.

        EXAMPLES::

            sage: f = ntl.ZZX([1,2,3,4,5])
            sage: f.truncate(3)
            [1 2 3]
            sage: f.truncate(8)
            [1 2 3 4 5]
            sage: f.truncate(1)
            [1]
            sage: f.truncate(0)
            []
            sage: f.truncate(-1)
            []
            sage: f.truncate(-5)
            []
        """
        if m <= 0:
            from copy import copy
            return copy(zero_ZZX)
        sig_on()
        return make_ZZX_sig_off(ZZX_truncate(&self.x, m))

    def multiply_and_truncate(self, ntl_ZZX other, long m):
        """
        Return self*other but with terms of degree >= m removed.

        EXAMPLES::

            sage: f = ntl.ZZX([1,2,3,4,5])
            sage: g = ntl.ZZX([10])
            sage: f.multiply_and_truncate(g, 2)
            [10 20]
            sage: g.multiply_and_truncate(f, 2)
            [10 20]
        """
        if m <= 0:
            from copy import copy
            return copy(zero_ZZX)
        return make_ZZX(ZZX_multiply_and_truncate(&self.x, &other.x, m))

    def square_and_truncate(self, long m):
        """
        Return self*self but with terms of degree >= m removed.

        EXAMPLES::

            sage: f = ntl.ZZX([1,2,3,4,5])
            sage: f.square_and_truncate(4)
            [1 4 10 20]
            sage: (f*f).truncate(4)
            [1 4 10 20]
        """
        if m < 0:
            from copy import copy
            return copy(zero_ZZX)
        return make_ZZX(ZZX_square_and_truncate(&self.x, m))

    def invert_and_truncate(self, long m):
        """
        Compute and return the inverse of ``self`` modulo `x^m`.
        The constant term of ``self`` must be 1 or -1.

        EXAMPLES::

            sage: f = ntl.ZZX([1,2,3,4,5,6,7])
            sage: f.invert_and_truncate(20)
            [1 -2 1 0 0 0 0 8 -23 22 -7 0 0 0 64 -240 337 -210 49]
            sage: g = f.invert_and_truncate(20)
            sage: g * f
            [1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 -512 1344 -1176 343]
        """
        if m < 0:
            raise ArithmeticError("m (=%s) must be positive" % m)
        n = self.constant_term()
        if n != ntl_ZZ(1) and n != ntl_ZZ(-1):
            raise ArithmeticError("The constant term of self must be 1 or -1.")
        sig_on()
        return make_ZZX_sig_off(ZZX_invert_and_truncate(&self.x, m))

    def multiply_mod(self, ntl_ZZX other, ntl_ZZX modulus):
        """
        Return ``self*other % modulus``.  The modulus must be monic with
        deg(modulus) > 0, and ``self`` and ``other`` must have smaller degree.

        EXAMPLES::

            sage: modulus = ntl.ZZX([1,2,0,1])    # must be monic
            sage: g = ntl.ZZX([-1,0,1])
            sage: h = ntl.ZZX([3,7,13])
            sage: h.multiply_mod(g, modulus)
            [-10 -34 -36]
        """
        sig_on()
        return make_ZZX_sig_off(ZZX_multiply_mod(&self.x, &other.x, &modulus.x))

    def trace_mod(self, ntl_ZZX modulus):
        """
        Return the trace of this polynomial modulus the modulus.
        The modulus must be monic, and of positive degree bigger
        than the degree of ``self``.

        EXAMPLES::

            sage: f = ntl.ZZX([1,2,0,3])
            sage: mod = ntl.ZZX([5,3,-1,1,1])
            sage: f.trace_mod(mod)
            -37
        """
        sig_on()
        return make_ZZ_sig_off(ZZX_trace_mod(&self.x, &modulus.x))

    def trace_list(self):
        """
        Return the list of traces of the powers `x^i` of the
        monomial x modulo this polynomial for i = 0, ..., deg(f)-1.
        This polynomial must be monic.

        EXAMPLES::

            sage: f = ntl.ZZX([1,2,0,3,0,1])
            sage: f.trace_list()
            [5, 0, -6, 0, 10]

        The input polynomial must be monic or a :exc:`ValueError` is raised::

            sage: f = ntl.ZZX([1,2,0,3,0,2])
            sage: f.trace_list()
            Traceback (most recent call last):
            ...
            ValueError: polynomial must be monic.
        """
        if not self.is_monic():
            raise ValueError("polynomial must be monic.")
        sig_on()
        cdef char* t
        t = ZZX_trace_list(&self.x)
        r = eval(char_to_str(t).replace(' ', ','))
        string_delete(t)
        return r

    def resultant(self, ntl_ZZX other, proof=None):
        """
        Return the resultant of ``self`` and ``other``.  If proof = False (the
        default is proof=None, see proof.polynomial or sage.structure.proof,
        but the global default is True), then this function may use a
        randomized strategy that errors with probability no more than
        `2^{-80}`.

        EXAMPLES::

            sage: f = ntl.ZZX([17,0,1,1])
            sage: g = ntl.ZZX([34,-17,18,2])
            sage: f.resultant(g)
            1345873
            sage: f.resultant(g, proof=False)
            1345873
        """
        proof = proof_flag(proof)
        # NOTE: Within a factor of 2 in speed compared to MAGMA.
        sig_on()
        return make_ZZ_sig_off(ZZX_resultant(&self.x, &other.x, proof))

    def norm_mod(self, ntl_ZZX modulus, proof=None):
        """
        Return the norm of this polynomial modulo the modulus.

        The modulus must be monic, and of positive degree strictly
        greater than the degree of ``self``.  If proof=False (the default
        is proof=None, see proof.polynomial or sage.structure.proof,
        but the global default is proof=True) then it may use a
        randomized strategy that errors with probability no more than
        `2^{-80}`.

        EXAMPLES::

            sage: f = ntl.ZZX([1,2,0,3])
            sage: mod = ntl.ZZX([-5,2,0,0,1])
            sage: f.norm_mod(mod)
            -8846

        The norm is the constant term of the characteristic polynomial::

            sage: f.charpoly_mod(mod)
            [-8846 -594 -60 14 1]
        """
        proof = proof_flag(proof)
        sig_on()
        return make_ZZ_sig_off(ZZX_norm_mod(&self.x, &modulus.x, proof))

    def discriminant(self, proof=None):
        r"""
        Return the discriminant of self, which is by definition
        $$
                (-1)^{m(m-1)/2} {\mbox{\tt resultant}}(a, a')/lc(a),
        $$
        where m = deg(a), and lc(a) is the leading coefficient of a.
        If proof is False (the default is proof=None, see
        proof.polynomial or sage.structure.proof, but the global
        default is proof=True), then this function may use a
        randomized strategy that errors with probability no more than
        `2^{-80}`.

        EXAMPLES::

            sage: f = ntl.ZZX([1,2,0,3])
            sage: f.discriminant()
            -339
            sage: f.discriminant(proof=False)
            -339
        """
        proof = proof_flag(proof)
        sig_on()
        return make_ZZ_sig_off(ZZX_discriminant(&self.x, proof))

    #def __call__(self, ntl_ZZ a):
    #    sig_on()
    #    return make_ZZ_sig_off(ZZX_polyeval(&self.x, a.x))

    def charpoly_mod(self, ntl_ZZX modulus, proof=None):
        """
        Return the characteristic polynomial of this polynomial modulo
        the modulus.  The modulus must be monic of degree bigger than
        ``self``. If proof=False (the default is proof=None, see
        proof.polynomial or sage.structure.proof, but the global
        default is proof=True), then this function may use a
        randomized strategy that errors with probability no more than
        `2^{-80}`.

        EXAMPLES::

            sage: f = ntl.ZZX([1,2,0,3])
            sage: mod = ntl.ZZX([-5,2,0,0,1])
            sage: f.charpoly_mod(mod)
            [-8846 -594 -60 14 1]
        """
        proof = proof_flag(proof)
        sig_on()
        return make_ZZX_sig_off(ZZX_charpoly_mod(&self.x, &modulus.x, proof))

    def minpoly_mod_noproof(self, ntl_ZZX modulus):
        """
        Return the minimal polynomial of this polynomial modulo the
        modulus.  The modulus must be monic of degree bigger than
        ``self``.  In all cases, this function may use a randomized
        strategy that errors with probability no more than `2^{-80}`.

        EXAMPLES::

            sage: f = ntl.ZZX([0,0,1])
            sage: g = f*f
            sage: f.charpoly_mod(g)
            [0 0 0 0 1]

        However, since `f^2 = 0` modulo `g`, its minimal polynomial
        is of degree `2`::

            sage: f.minpoly_mod_noproof(g)
            [0 0 1]
        """
        sig_on()
        return make_ZZX_sig_off(ZZX_minpoly_mod(&self.x, &modulus.x))

    def clear(self):
        """
        Reset this polynomial to 0.  Changes this polynomial in place.

        EXAMPLES::

            sage: f = ntl.ZZX([1,2,3])
            sage: f
            [1 2 3]
            sage: f.clear()
            sage: f
            []
        """
        ZZX_clear(&self.x)

    def preallocate_space(self, long n):
        """
        Pre-allocate spaces for n coefficients.  The polynomial that f
        represents is unchanged.  This is useful if you know you'll be
        setting coefficients up to n, so memory isn't re-allocated as
        the polynomial grows.  (You might save a millisecond with this
        function.)

        EXAMPLES::

            sage: f = ntl.ZZX([1,2,3])
            sage: f.preallocate_space(20)
            sage: f
            [1 2 3]
            sage: f[10]=5  # no new memory is allocated
            sage: f
            [1 2 3 0 0 0 0 0 0 0 5]
        """
        sig_on()
        ZZX_preallocate_space(&self.x, n)
        sig_off()

    def squarefree_decomposition(self):
        """
        Return the square-free decomposition of ``self`` (a partial
        factorization into square-free, relatively prime polynomials)
        as a list of 2-tuples, where the first element in each tuple
        is a factor, and the second is its exponent.
        Assumes that ``self`` is primitive.

        EXAMPLES::

            sage: f = ntl.ZZX([0, 1, 2, 1])
            sage: f.squarefree_decomposition()
            [([0 1], 1), ([1 1], 2)]
        """
        cdef ZZX_c** v
        cdef long* e
        cdef long i, n
        sig_on()
        ZZX_squarefree_decomposition(&v, &e, &n, &self.x)
        sig_off()
        F = []
        for i from 0 <= i < n:
            F.append((make_ZZX(v[i]), e[i]))
        sig_free(v)
        sig_free(e)
        return F


one_ZZX = ntl_ZZX([1])
zero_ZZX = ntl_ZZX()



# Automatically generated by /build/sagemath/src/sage/src/sage_setup/autogen/interpreters/internal/generator.py.  Do not edit!
cimport sage.libs.gsl.math  # Add dependency on GSL

from cpython.ref cimport PyObject
cdef extern from "Python.h":
    void Py_DECREF(PyObject *o)
    void Py_INCREF(PyObject *o)
    void Py_CLEAR(PyObject *o)

    object PyList_New(Py_ssize_t len)
    ctypedef struct PyListObject:
        PyObject **ob_item

    ctypedef struct PyTupleObject:
        PyObject **ob_item

from cysignals.memory cimport check_allocarray, sig_free

from sage.ext.fast_callable cimport Wrapper

cdef extern from "interp_rdf.c":
    double interp_rdf(double* args,
        double* constants,
        PyObject** py_constants,
        double* stack,
        int* code) except? -1094648009105371

cdef class Wrapper_rdf(Wrapper):
    # attributes are declared in corresponding .pxd file

    def __init__(self, args):
        Wrapper.__init__(self, args, metadata)
        cdef int i
        cdef int count
        count = args['args']
        self._n_args = count
        self._args = <double*>check_allocarray(self._n_args, sizeof(double))
        val = args['constants']
        self._n_constants = len(val)
        self._constants = <double*>check_allocarray(self._n_constants, sizeof(double))
        for i in range(len(val)):
            self._constants[i] = val[i]
        val = args['py_constants']
        self._n_py_constants = len(val)
        self._list_py_constants = PyList_New(self._n_py_constants)
        self._py_constants = (<PyListObject *>self._list_py_constants).ob_item
        for i in range(len(val)):
            self._py_constants[i] = <PyObject *>val[i]; Py_INCREF(self._py_constants[i])
        count = args['stack']
        self._n_stack = count
        self._stack = <double*>check_allocarray(self._n_stack, sizeof(double))
        val = args['code']
        self._n_code = len(val)
        self._code = <int*>check_allocarray(self._n_code, sizeof(int))
        for i in range(len(val)):
            self._code[i] = val[i]
        self._domain = args['domain']

    def __dealloc__(self):
        cdef int i
        if self._args:
            sig_free(self._args)
        if self._constants:
            sig_free(self._constants)
        if self._stack:
            sig_free(self._stack)
        if self._code:
            sig_free(self._code)

    def __call__(self, *args):
        if self._n_args != len(args): raise ValueError
        cdef double* c_args = self._args
        cdef int i
        for i from 0 <= i < len(args):
            self._args[i] = args[i]
        return self._domain(interp_rdf(c_args
            , self._constants
            , self._py_constants
            , self._stack
            , self._code
            ))

    cdef bint call_c(self,
                     double* args,
                     double* result) except 0:
        result[0] = interp_rdf(args
            , self._constants
            , self._py_constants
            , self._stack
            , self._code
            )

        return 1

from sage.ext.fast_callable import CompilerInstrSpec, InterpreterMetadata
metadata = InterpreterMetadata(by_opname={
  'load_arg':
  (CompilerInstrSpec(0, 1, ['args']), 0),
  'load_const':
  (CompilerInstrSpec(0, 1, ['constants']), 1),
  'return':
  (CompilerInstrSpec(1, 0, []), 2),
  'py_call':
  (CompilerInstrSpec(0, 1, ['py_constants', 'n_inputs']), 3),
  'pow':
  (CompilerInstrSpec(2, 1, []), 4),
  'add':
  (CompilerInstrSpec(2, 1, []), 5),
  'sub':
  (CompilerInstrSpec(2, 1, []), 6),
  'mul':
  (CompilerInstrSpec(2, 1, []), 7),
  'div':
  (CompilerInstrSpec(2, 1, []), 8),
  'ipow':
  (CompilerInstrSpec(1, 1, ['code']), 9),
  'neg':
  (CompilerInstrSpec(1, 1, []), 10),
  'invert':
  (CompilerInstrSpec(1, 1, []), 11),
  'abs':
  (CompilerInstrSpec(1, 1, []), 12),
  'sqrt':
  (CompilerInstrSpec(1, 1, []), 13),
  'ceil':
  (CompilerInstrSpec(1, 1, []), 14),
  'floor':
  (CompilerInstrSpec(1, 1, []), 15),
  'sin':
  (CompilerInstrSpec(1, 1, []), 16),
  'cos':
  (CompilerInstrSpec(1, 1, []), 17),
  'tan':
  (CompilerInstrSpec(1, 1, []), 18),
  'asin':
  (CompilerInstrSpec(1, 1, []), 19),
  'acos':
  (CompilerInstrSpec(1, 1, []), 20),
  'atan':
  (CompilerInstrSpec(1, 1, []), 21),
  'sinh':
  (CompilerInstrSpec(1, 1, []), 22),
  'cosh':
  (CompilerInstrSpec(1, 1, []), 23),
  'tanh':
  (CompilerInstrSpec(1, 1, []), 24),
  'asinh':
  (CompilerInstrSpec(1, 1, []), 25),
  'acosh':
  (CompilerInstrSpec(1, 1, []), 26),
  'atanh':
  (CompilerInstrSpec(1, 1, []), 27),
  'exp':
  (CompilerInstrSpec(1, 1, []), 28),
  'log':
  (CompilerInstrSpec(1, 1, []), 29),
 },
 by_opcode=[
  ('load_arg',
   CompilerInstrSpec(0, 1, ['args'])),
  ('load_const',
   CompilerInstrSpec(0, 1, ['constants'])),
  ('return',
   CompilerInstrSpec(1, 0, [])),
  ('py_call',
   CompilerInstrSpec(0, 1, ['py_constants', 'n_inputs'])),
  ('pow',
   CompilerInstrSpec(2, 1, [])),
  ('add',
   CompilerInstrSpec(2, 1, [])),
  ('sub',
   CompilerInstrSpec(2, 1, [])),
  ('mul',
   CompilerInstrSpec(2, 1, [])),
  ('div',
   CompilerInstrSpec(2, 1, [])),
  ('ipow',
   CompilerInstrSpec(1, 1, ['code'])),
  ('neg',
   CompilerInstrSpec(1, 1, [])),
  ('invert',
   CompilerInstrSpec(1, 1, [])),
  ('abs',
   CompilerInstrSpec(1, 1, [])),
  ('sqrt',
   CompilerInstrSpec(1, 1, [])),
  ('ceil',
   CompilerInstrSpec(1, 1, [])),
  ('floor',
   CompilerInstrSpec(1, 1, [])),
  ('sin',
   CompilerInstrSpec(1, 1, [])),
  ('cos',
   CompilerInstrSpec(1, 1, [])),
  ('tan',
   CompilerInstrSpec(1, 1, [])),
  ('asin',
   CompilerInstrSpec(1, 1, [])),
  ('acos',
   CompilerInstrSpec(1, 1, [])),
  ('atan',
   CompilerInstrSpec(1, 1, [])),
  ('sinh',
   CompilerInstrSpec(1, 1, [])),
  ('cosh',
   CompilerInstrSpec(1, 1, [])),
  ('tanh',
   CompilerInstrSpec(1, 1, [])),
  ('asinh',
   CompilerInstrSpec(1, 1, [])),
  ('acosh',
   CompilerInstrSpec(1, 1, [])),
  ('atanh',
   CompilerInstrSpec(1, 1, [])),
  ('exp',
   CompilerInstrSpec(1, 1, [])),
  ('log',
   CompilerInstrSpec(1, 1, [])),
 ],
 ipow_range=(-2147483648, 2147483647))
